"""
    :codeauthor: Mike Wiebe <@mikewiebe>
"""

# Copyright (c) 2019 Cisco and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import salt.states.nxos as nxos_state
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {nxos_state: {}}


def test_user_present_create():
    """
    user_present method - create
    """

    roles = ["vdc-admin"]

    salt_mock = {
        "nxos.get_roles": MagicMock(side_effect=[[], roles, roles]),
        "nxos.get_user": MagicMock(side_effect=[""]),
        "nxos.set_role": MagicMock(side_effect=["set_role"]),
    }
    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.user_present("daniel", roles=roles)

            assert result["name"] == "daniel"
            assert result["result"]
            assert result["changes"]["roles"]["new"] == ["vdc-admin"]
            assert result["changes"]["roles"]["old"] == []
            assert result["comment"] == "User set correctly"


def test_user_present_create_opts_test():
    """
    user_present method - create opts
    """

    roles = ["vdc-admin"]

    salt_mock = {
        "nxos.get_roles": MagicMock(side_effect=[[], roles, roles]),
        "nxos.get_user": MagicMock(side_effect=[""]),
        "nxos.set_role": MagicMock(side_effect=["set_role"]),
    }
    with patch.dict(nxos_state.__opts__, {"test": True}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.user_present("daniel", roles=roles)

            assert result["name"] == "daniel"
            assert result["result"] is None
            assert result["changes"]["role"]["add"] == ["vdc-admin"]
            assert result["changes"]["role"]["remove"] == []
            assert result["comment"] == "User will be created"


def test_user_present_create_non_defaults():
    """
    user_present method - create non default opts
    """

    username = "daniel"
    password = "ghI&435y55#"
    roles = ["vdc-admin", "dev-ops"]
    encrypted = False
    crypt_salt = "foobar123"
    algorithm = "md5"

    salt_mock = {
        "nxos.check_password": MagicMock(side_effect=[False, True]),
        "nxos.get_roles": MagicMock(side_effect=[[], roles, roles]),
        "nxos.get_user": MagicMock(side_effect=[""]),
        "nxos.set_password": MagicMock(side_effect=["new_user"]),
        "nxos.set_role": MagicMock(side_effect=["set_role", "set_role"]),
    }
    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, salt_mock):
            result = nxos_state.user_present(
                username,
                password=password,
                roles=roles,
                encrypted=encrypted,
                crypt_salt=crypt_salt,
                algorithm=algorithm,
            )

            assert result["name"] == "daniel"
            assert result["result"]
            assert result["changes"]["password"]["new"] == "new_user"
            assert result["changes"]["password"]["old"] == ""
            assert result["changes"]["roles"]["new"] == ["vdc-admin", "dev-ops"]
            assert result["changes"]["roles"]["old"] == []
            assert result["comment"] == "User set correctly"


def test_user_present_create_encrypted_password_no_roles_opts_test():
    """
    user_present method - encrypted password, no roles
    """

    username = "daniel"
    password = "$1$foobar12$K7x4Rxua11qakvrRjcwDC/"
    encrypted = True
    crypt_salt = "foobar123"
    algorithm = "md5"

    salt_mock = {
        "nxos.check_password": MagicMock(side_effect=[False, True]),
        "nxos.get_user": MagicMock(side_effect=[""]),
        "nxos.set_password": MagicMock(side_effect=["new_user"]),
    }

    with patch.dict(nxos_state.__opts__, {"test": True}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.user_present(
                username,
                password=password,
                encrypted=encrypted,
                crypt_salt=crypt_salt,
                algorithm=algorithm,
            )

            assert result["name"] == "daniel"
            assert result["result"] is None
            assert result["changes"]["password"] is True
            assert result["comment"] == "User will be created"


def test_user_present_create_user_exists():
    """
    user_present method - user exists
    """

    username = "daniel"
    password = "$1$foobar12$K7x4Rxua11qakvrRjcwDC/"
    encrypted = True
    crypt_salt = "foobar123"
    algorithm = "md5"

    salt_mock = {
        "nxos.check_password": MagicMock(side_effect=[True]),
        "nxos.get_user": MagicMock(side_effect=["user_exists"]),
    }

    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.user_present(
                username,
                password=password,
                encrypted=encrypted,
                crypt_salt=crypt_salt,
                algorithm=algorithm,
            )

            assert result["name"] == "daniel"
            assert result["result"]
            assert result["changes"] == {}
            assert result["comment"] == "User already exists"


def test_user_present_create_user_exists_opts_test():
    """
    user_present method - user exists with opts
    """

    username = "daniel"
    password = "$1$foobar12$K7x4Rxua11qakvrRjcwDC/"
    roles = ["vdc-admin", "dev-opts"]
    new_roles = ["network-operator"]
    encrypted = True
    crypt_salt = "foobar123"
    algorithm = "md5"

    salt_mock = {
        "nxos.check_password": MagicMock(side_effect=[True]),
        "nxos.get_roles": MagicMock(side_effect=[roles]),
        "nxos.get_user": MagicMock(side_effect=["user_exists"]),
    }

    with patch.dict(nxos_state.__opts__, {"test": True}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.user_present(
                username,
                password=password,
                roles=new_roles,
                encrypted=encrypted,
                crypt_salt=crypt_salt,
                algorithm=algorithm,
            )

            remove = result["changes"]["roles"]["remove"]
            remove.sort()
            assert result["name"] == "daniel"
            assert result["result"] is None
            assert result["changes"]["roles"]["add"] == ["network-operator"]
            assert remove == ["dev-opts", "vdc-admin"]
            assert result["comment"] == "User will be updated"


def test_user_absent():
    """
    user_absent method - remove user
    """

    username = "daniel"

    salt_mock = {
        "nxos.get_user": MagicMock(side_effect=["daniel", ""]),
        "nxos.remove_user": MagicMock(side_effect=["remove_user"]),
    }

    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.user_absent(username)

            assert result["name"] == "daniel"
            assert result["result"]
            assert result["changes"]["old"] == "daniel"
            assert result["changes"]["new"] == ""
            assert result["comment"] == "User removed"


def test_user_absent_user_does_not_exist():
    """
    user_absent method - remove user
    """

    username = "daniel"

    side_effect = MagicMock(side_effect=[""])

    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, {"nxos.get_user": side_effect}):

            result = nxos_state.user_absent(username)

            assert result["name"] == "daniel"
            assert result["result"]
            assert result["changes"] == {}
            assert result["comment"] == "User does not exist"


def test_user_absent_test_opts():
    """
    user_absent method - remove user with opts
    """

    username = "daniel"

    salt_mock = {
        "nxos.get_user": MagicMock(side_effect=["daniel", ""]),
        "nxos.remove_user": MagicMock(side_effect=["remove_user"]),
    }

    with patch.dict(nxos_state.__opts__, {"test": True}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.user_absent(username)

            assert result["name"] == "daniel"
            assert result["result"] is None
            assert result["changes"]["old"] == "daniel"
            assert result["changes"]["new"] == ""
            assert result["comment"] == "User will be removed"


def test_config_present():
    """
    config_present method - add config
    """

    config_data = [
        "snmp-server community randomSNMPstringHERE group network-operator",
        "snmp-server community AnotherRandomSNMPSTring group network-admin",
    ]
    snmp_matches1 = [
        "snmp-server community randomSNMPstringHERE group network-operator"
    ]
    snmp_matches2 = [
        ["snmp-server community AnotherRandomSNMPSTring group network-admin"]
    ]

    salt_mock = {
        "nxos.config": MagicMock(side_effect=["add_snmp_config1", "add_snmp_config2"]),
        "nxos.find": MagicMock(side_effect=[[], snmp_matches1, snmp_matches2]),
    }

    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.config_present(config_data)

            assert result["name"] == config_data
            assert result["result"]
            assert result["changes"]["new"] == config_data
            assert result["comment"] == "Successfully added config"


def test_config_present_already_configured():
    """
    config_present method - add config already configured
    """

    config_data = [
        "snmp-server community randomSNMPstringHERE group network-operator",
        "snmp-server community AnotherRandomSNMPSTring group network-admin",
    ]

    side_effect = MagicMock(side_effect=[config_data[0], config_data[1]])

    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, {"nxos.find": side_effect}):

            result = nxos_state.config_present(config_data)

            assert result["name"] == config_data
            assert result["result"]
            assert result["changes"] == {}
            assert result["comment"] == "Config is already set"


def test_config_present_test_opts():
    """
    config_present method - add config
    """

    config_data = [
        "snmp-server community randomSNMPstringHERE group network-operator",
        "snmp-server community AnotherRandomSNMPSTring group network-admin",
    ]
    snmp_matches1 = [
        "snmp-server community randomSNMPstringHERE group network-operator"
    ]
    snmp_matches2 = [
        ["snmp-server community AnotherRandomSNMPSTring group network-admin"]
    ]

    salt_mock = {
        "nxos.config": MagicMock(side_effect=["add_snmp_config1", "add_snmp_config2"]),
        "nxos.find": MagicMock(side_effect=[[], snmp_matches1, snmp_matches2]),
    }

    with patch.dict(nxos_state.__opts__, {"test": True}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.config_present(config_data)

            assert result["name"] == config_data
            assert result["result"] is None
            assert result["changes"]["new"] == config_data
            assert result["comment"] == "Config will be added"


def test_config_present_fail_to_add():
    """
    config_present method - add config fails
    """

    config_data = [
        "snmp-server community randomSNMPstringHERE group network-operator",
        "snmp-server community AnotherRandomSNMPSTring group network-admin",
    ]
    snmp_matches1 = [
        "snmp-server community randomSNMPstringHERE group network-operator"
    ]
    snmp_matches2 = [
        ["snmp-server community AnotherRandomSNMPSTring group network-admin"]
    ]

    salt_mock = {
        "nxos.config": MagicMock(side_effect=["add_snmp_config1", "add_snmp_config2"]),
        "nxos.find": MagicMock(side_effect=[[], "", ""]),
    }

    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.config_present(config_data)

            assert result["name"] == config_data
            assert not result["result"]
            assert result["changes"] == {}
            assert result["comment"] == "Failed to add config"


def test_replace():
    """
    replace method - replace config
    """

    name = "randomSNMPstringHERE"
    repl = "NEWrandoSNMPstringHERE"
    matches_before = [
        "snmp-server community randomSNMPstringHERE group network-operator"
    ]
    match_after = []
    changes = {}
    changes["new"] = [
        "snmp-server community NEWrandoSNMPstringHERE group network-operator"
    ]
    changes["old"] = [
        "snmp-server community randomSNMPstringHERE group network-operator"
    ]

    salt_mock = {
        "nxos.find": MagicMock(side_effect=[matches_before, match_after]),
        "nxos.replace": MagicMock(side_effect=[changes]),
    }

    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.replace(name, repl)

            assert result["name"] == name
            assert result["result"]
            assert result["changes"]["new"] == changes["new"]
            assert result["changes"]["old"] == changes["old"]
            assert (
                result["comment"]
                == 'Successfully replaced all instances of "randomSNMPstringHERE" with'
                ' "NEWrandoSNMPstringHERE"'
            )


def test_replace_test_opts():
    """
    replace method - replace config
    """

    name = "randomSNMPstringHERE"
    repl = "NEWrandoSNMPstringHERE"
    matches_before = [
        "snmp-server community randomSNMPstringHERE group network-operator"
    ]
    match_after = []
    changes = {}
    changes["new"] = [
        "snmp-server community NEWrandoSNMPstringHERE group network-operator"
    ]
    changes["old"] = [
        "snmp-server community randomSNMPstringHERE group network-operator"
    ]

    salt_mock = {
        "nxos.find": MagicMock(side_effect=[matches_before, match_after]),
        "nxos.replace": MagicMock(side_effect=[changes]),
    }

    with patch.dict(nxos_state.__opts__, {"test": True}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.replace(name, repl)

            assert result["name"] == name
            assert result["result"] is None
            assert result["changes"]["new"] == changes["new"]
            assert result["changes"]["old"] == changes["old"]
            assert result["comment"] == "Configs will be changed"


def test_config_absent():
    """
    config_absent method - remove config
    """

    config_data = [
        "snmp-server community randomSNMPstringHERE group network-operator",
        "snmp-server community AnotherRandomSNMPSTring group network-admin",
    ]
    snmp_matches1 = [
        "snmp-server community randomSNMPstringHERE group network-operator"
    ]
    snmp_matches2 = [
        ["snmp-server community AnotherRandomSNMPSTring group network-admin"]
    ]

    salt_mock = {
        "nxos.delete_config": MagicMock(side_effect=["remove_config", "remove_config"]),
        "nxos.find": MagicMock(side_effect=[snmp_matches1, [], snmp_matches2, []]),
    }

    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.config_absent(config_data)

            assert result["name"] == config_data
            assert result["result"]
            assert result["changes"]["new"] == config_data
            assert result["comment"] == "Successfully deleted config"


def test_config_absent_already_configured():
    """
    config_absent method - add config removed
    """

    config_data = [
        "snmp-server community randomSNMPstringHERE group network-operator",
        "snmp-server community AnotherRandomSNMPSTring group network-admin",
    ]

    side_effect = MagicMock(side_effect=[[], []])

    with patch.dict(nxos_state.__opts__, {"test": False}):
        with patch.dict(nxos_state.__salt__, {"nxos.find": side_effect}):

            result = nxos_state.config_absent(config_data)

            assert result["name"] == config_data
            assert result["result"]
            assert result["changes"] == {}
            assert result["comment"] == "Config is already absent"


def test_config_absent_test_opts():
    """
    config_absent method - remove config
    """

    config_data = [
        "snmp-server community randomSNMPstringHERE group network-operator",
        "snmp-server community AnotherRandomSNMPSTring group network-admin",
    ]
    snmp_matches1 = [
        "snmp-server community randomSNMPstringHERE group network-operator"
    ]
    snmp_matches2 = [
        ["snmp-server community AnotherRandomSNMPSTring group network-admin"]
    ]

    salt_mock = {
        "nxos.delete_config": MagicMock(side_effect=["remove_config", "remove_config"]),
        "nxos.find": MagicMock(side_effect=[snmp_matches1, [], snmp_matches2, []]),
    }

    with patch.dict(nxos_state.__opts__, {"test": True}):
        with patch.dict(nxos_state.__salt__, salt_mock):

            result = nxos_state.config_absent(config_data)

            assert result["name"] == config_data
            assert result["result"] is None
            assert result["changes"]["new"] == config_data
            assert result["comment"] == "Config will be removed"



import tempfile

import jinja2.exceptions
import pytest

import salt.modules.debian_ip as debian_ip
import salt.utils.files
from tests.support.mock import MagicMock, patch

try:
    from collections import OrderedDict as odict
except ImportError:
    from collections import OrderedDict as odict

# Big pile of interface data for unit tests
#   To skip, search for 'DebianIpTestCase'
# fmt: off


pytestmark = [
    pytest.mark.skip_on_windows(reason="Do not run these tests on Windows"),
    pytest.mark.skip_on_darwin(reason="Do not run these tests on Mac"),
]


@pytest.fixture
def test_interfaces():
    return [
        # Structure
        #{'iface_name': 'ethX', 'iface_type': 'eth', 'enabled': True,
        #    'skip_test': bool(),        # True to disable this test
        #    'build_interface': dict(),  # data read from sls
        #    'get_interface(): OrderedDict(),   # data read from interfaces file
        #    'return': list()},          # jinja-rendered data

        # IPv4-only interface; single address
        {'iface_name': 'eth1', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'ipaddr': '192.168.4.9',
                'netmask': '255.255.255.0',
                'gateway': '192.168.4.1',
                'enable_ipv6': False,
                'noifupdown': True,
                },
            'get_interface': odict([('eth1', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '192.168.4.9'),
                    ('netmask', '255.255.255.0'),
                    ('gateway', '192.168.4.1'),
                    ])),
                ]))]))]),
            'return': [
                'auto eth1\n',
                'iface eth1 inet static\n',
                '    address 192.168.4.9\n',
                '    netmask 255.255.255.0\n',
                '    gateway 192.168.4.1\n',
                '\n']},

        # IPv6-only; single address
        {'iface_name': 'eth2', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'ipv6proto': 'static',
                'ipv6ipaddr': '2001:db8:dead:beef::3',
                'ipv6netmask': '64',
                'ipv6gateway': '2001:db8:dead:beef::1',
                'enable_ipv6': True,
                'noifupdown': True,
                },
            'get_interface': odict([('eth2', odict([('enabled', True), ('data', odict([
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '2001:db8:dead:beef::3'),
                    ('netmask', 64),
                    ('gateway', '2001:db8:dead:beef::1'),
                    ])),
                ]))]))]),
            'return': [
                'auto eth2\n',
                'iface eth2 inet6 static\n',
                '    address 2001:db8:dead:beef::3\n',
                '    netmask 64\n',
                '    gateway 2001:db8:dead:beef::1\n',
                '\n']},

        # IPv6-only; multiple addrs; no gw; first addr from ipv6addr
        {'iface_name': 'eth3', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'ipv6proto': 'static',
                'ipv6ipaddr': '2001:db8:dead:beef::5/64',
                'ipv6ipaddrs': [
                    '2001:db8:dead:beef::7/64',
                    '2001:db8:dead:beef::8/64',
                    '2001:db8:dead:beef::9/64'],
                'enable_ipv6': True,
                'noifupdown': True,
                },
            'get_interface': odict([('eth3', odict([('enabled', True), ('data', odict([
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '2001:db8:dead:beef::5/64'),
                    ('addresses', [
                        '2001:db8:dead:beef::7/64',
                        '2001:db8:dead:beef::8/64',
                        '2001:db8:dead:beef::9/64',
                        ]),
                    ])),
                ]))]))]),
            'return': [
                'auto eth3\n',
                'iface eth3 inet6 static\n',
                '    address 2001:db8:dead:beef::5/64\n',
                '    address 2001:db8:dead:beef::7/64\n',
                '    address 2001:db8:dead:beef::8/64\n',
                '    address 2001:db8:dead:beef::9/64\n',
                '\n']},

        # IPv6-only; multiple addresses
        {'iface_name': 'eth4', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'ipv6proto': 'static',
                'ipv6ipaddrs': [
                    '2001:db8:dead:beef::5/64',
                    '2001:db8:dead:beef::7/64',
                    '2001:db8:dead:beef::8/64',
                    '2001:db8:dead:beef::9/64'],
                'ipv6gateway': '2001:db8:dead:beef::1',
                'enable_ipv6': True,
                'noifupdown': True,
                },
            'get_interface': odict([('eth4', odict([('enabled', True), ('data', odict([
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '2001:db8:dead:beef::5/64'),
                    ('addresses', [
                        '2001:db8:dead:beef::7/64',
                        '2001:db8:dead:beef::8/64',
                        '2001:db8:dead:beef::9/64',
                        ]),
                    ('gateway', '2001:db8:dead:beef::1'),
                    ])),
                ]))]))]),
            'return': [
                'auto eth4\n',
                'iface eth4 inet6 static\n',
                '    address 2001:db8:dead:beef::5/64\n',
                '    address 2001:db8:dead:beef::7/64\n',
                '    address 2001:db8:dead:beef::8/64\n',
                '    address 2001:db8:dead:beef::9/64\n',
                '    gateway 2001:db8:dead:beef::1\n',
                '\n']},

        # IPv4 and IPv6 settings with v4 disabled
        {'iface_name': 'eth5', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'ipaddr': '192.168.4.9',
                'netmask': '255.255.255.0',
                'gateway': '192.168.4.1',
                'ipv6proto': 'static',
                'ipv6ipaddr': '2001:db8:dead:beef::3',
                'ipv6netmask': '64',
                'ipv6gateway': '2001:db8:dead:beef::1',
                'enable_ipv4': False,
                'noifupdown': True,
                },
            'get_interface': odict([('eth5', odict([('enabled', True), ('data', odict([
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '2001:db8:dead:beef::3'),
                    ('netmask', 64),
                    ('gateway', '2001:db8:dead:beef::1'),
                    ])),
                ]))]))]),
            'return': [
                'auto eth5\n',
                'iface eth5 inet6 static\n',
                '    address 2001:db8:dead:beef::3\n',
                '    netmask 64\n',
                '    gateway 2001:db8:dead:beef::1\n',
                '\n']},

        # IPv4 and IPv6 settings with v6 disabled
        {'iface_name': 'eth6', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'ipaddr': '192.168.4.9',
                'netmask': '255.255.255.0',
                'gateway': '192.168.4.1',
                'ipv6proto': 'static',
                'ipv6ipaddr': '2001:db8:dead:beef::3',
                'ipv6netmask': '64',
                'ipv6gateway': '2001:db8:dead:beef::1',
                'enable_ipv6': False,
                'noifupdown': True,
                },
            'get_interface': odict([('eth6', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '192.168.4.9'),
                    ('netmask', '255.255.255.0'),
                    ('gateway', '192.168.4.1'),
                    ])),
                ]))]))]),
            'return': [
                'auto eth6\n',
                'iface eth6 inet static\n',
                '    address 192.168.4.9\n',
                '    netmask 255.255.255.0\n',
                '    gateway 192.168.4.1\n',
                '\n']},

        # IPv4 and IPv6; shared/overridden settings
        {'iface_name': 'eth7', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'ipaddr': '192.168.4.9',
                'netmask': '255.255.255.0',
                'gateway': '192.168.4.1',
                'ipv6proto': 'static',
                'ipv6ipaddr': '2001:db8:dead:beef::3',
                'ipv6netmask': '64',
                'ipv6gateway': '2001:db8:dead:beef::1',
                'ttl': '18',  # shared
                'ipv6ttl': '15',  # overridden for v6
                'mtu': '1480',  # shared
                'enable_ipv6': True,
                'noifupdown': True,
                },
            'get_interface': odict([('eth7', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '192.168.4.9'),
                    ('netmask', '255.255.255.0'),
                    ('gateway', '192.168.4.1'),
                    ('ttl', 18),
                    ('mtu', 1480),
                    ])),
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '2001:db8:dead:beef::3'),
                    ('netmask', 64),
                    ('gateway', '2001:db8:dead:beef::1'),
                    ('ttl', 15),
                    ('mtu', 1480),
                    ])),
                ]))]))]),
            'return': [
                'auto eth7\n',
                'iface eth7 inet static\n',
                '    address 192.168.4.9\n',
                '    netmask 255.255.255.0\n',
                '    gateway 192.168.4.1\n',
                '    ttl 18\n',
                '    mtu 1480\n',
                'iface eth7 inet6 static\n',
                '    address 2001:db8:dead:beef::3\n',
                '    netmask 64\n',
                '    gateway 2001:db8:dead:beef::1\n',
                '    ttl 15\n',
                '    mtu 1480\n',
                '\n']},

        # Slave iface
        {'iface_name': 'eth8', 'iface_type': 'slave', 'enabled': True,
            'build_interface': {
                'master': 'bond0',
                'noifupdown': True,
                },
            'get_interface': odict([('eth8', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'manual'),
                    ('filename', None),
                    ('bonding', odict([
                        ('master', 'bond0'),
                        ])),
                    ('bonding_keys', ['master']),
                    ])),
                ]))]))]),
            'return': [
                'auto eth8\n',
                'iface eth8 inet manual\n',
                '    bond-master bond0\n',
                '\n']},

        # Bond; with address IPv4 and IPv6 address; slaves as string
        {'iface_name': 'bond9', 'iface_type': 'bond', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'ipaddr': '10.1.0.14',
                'netmask': '255.255.255.0',
                'gateway': '10.1.0.1',
                'ipv6proto': 'static',
                'ipv6ipaddr': '2001:db8:dead:c0::3',
                'ipv6netmask': '64',
                'ipv6gateway': '2001:db8:dead:c0::1',
                'mode': '802.3ad',
                'slaves': 'eth4 eth5',
                'enable_ipv6': True,
                'noifupdown': True,
                },
            'get_interface': odict([('bond9', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '10.1.0.14'),
                    ('netmask', '255.255.255.0'),
                    ('gateway', '10.1.0.1'),
                    ('bonding', odict([
                        ('ad_select', '0'),
                        ('downdelay', '200'),
                        ('lacp_rate', '0'),
                        ('miimon', '100'),
                        ('mode', '4'),
                        ('slaves', 'eth4 eth5'),
                        ('updelay', '0'),
                        ('use_carrier', 'on'),
                        ])),
                    ('bonding_keys', [
                        'ad_select',
                        'downdelay',
                        'lacp_rate',
                        'miimon',
                        'mode',
                        'slaves',
                        'updelay',
                        'use_carrier',
                        ]),
                    ])),
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '2001:db8:dead:c0::3'),
                    ('netmask', 64),
                    ('gateway', '2001:db8:dead:c0::1'),
                    ('bonding', odict([
                        ('ad_select', '0'),
                        ('downdelay', '200'),
                        ('lacp_rate', '0'),
                        ('miimon', '100'),
                        ('mode', '4'),
                        ('slaves', 'eth4 eth5'),
                        ('updelay', '0'),
                        ('use_carrier', 'on'),
                        ])),
                    ('bonding_keys', [
                        'ad_select',
                        'downdelay',
                        'lacp_rate',
                        'miimon',
                        'mode',
                        'slaves',
                        'updelay',
                        'use_carrier',
                        ]),
                    ])),
                ]))]))]),
            'return': [
                'auto bond9\n',
                'iface bond9 inet static\n',
                '    address 10.1.0.14\n',
                '    netmask 255.255.255.0\n',
                '    gateway 10.1.0.1\n',
                '    bond-ad_select 0\n',
                '    bond-downdelay 200\n',
                '    bond-lacp_rate 0\n',
                '    bond-miimon 100\n',
                '    bond-mode 4\n',
                '    bond-slaves eth4 eth5\n',
                '    bond-updelay 0\n',
                '    bond-use_carrier on\n',
                'iface bond9 inet6 static\n',
                '    address 2001:db8:dead:c0::3\n',
                '    netmask 64\n',
                '    gateway 2001:db8:dead:c0::1\n',
                '    bond-ad_select 0\n',
                '    bond-downdelay 200\n',
                '    bond-lacp_rate 0\n',
                '    bond-miimon 100\n',
                '    bond-mode 4\n',
                '    bond-slaves eth4 eth5\n',
                '    bond-updelay 0\n',
                '    bond-use_carrier on\n',
                '\n']},

        # Bond; with address IPv4 and IPv6 address; slaves as list
        {'iface_name': 'bond10', 'iface_type': 'bond', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'ipaddr': '10.1.0.14',
                'netmask': '255.255.255.0',
                'gateway': '10.1.0.1',
                'ipv6proto': 'static',
                'ipv6ipaddr': '2001:db8:dead:c0::3',
                'ipv6netmask': '64',
                'ipv6gateway': '2001:db8:dead:c0::1',
                'mode': '802.3ad',
                'slaves': ['eth4', 'eth5'],
                'enable_ipv6': True,
                'noifupdown': True,
                },
            'get_interface': odict([('bond10', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '10.1.0.14'),
                    ('netmask', '255.255.255.0'),
                    ('gateway', '10.1.0.1'),
                    ('bonding', odict([
                        ('ad_select', '0'),
                        ('downdelay', '200'),
                        ('lacp_rate', '0'),
                        ('miimon', '100'),
                        ('mode', '4'),
                        ('slaves', 'eth4 eth5'),
                        ('updelay', '0'),
                        ('use_carrier', 'on'),
                        ])),
                    ('bonding_keys', [
                        'ad_select',
                        'downdelay',
                        'lacp_rate',
                        'miimon',
                        'mode',
                        'slaves',
                        'updelay',
                        'use_carrier',
                        ]),
                    ])),
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '2001:db8:dead:c0::3'),
                    ('netmask', 64),
                    ('gateway', '2001:db8:dead:c0::1'),
                    ('bonding', odict([
                        ('ad_select', '0'),
                        ('downdelay', '200'),
                        ('lacp_rate', '0'),
                        ('miimon', '100'),
                        ('mode', '4'),
                        ('slaves', 'eth4 eth5'),
                        ('updelay', '0'),
                        ('use_carrier', 'on'),
                        ])),
                    ('bonding_keys', [
                        'ad_select',
                        'downdelay',
                        'lacp_rate',
                        'miimon',
                        'mode',
                        'slaves',
                        'updelay',
                        'use_carrier',
                        ]),
                    ])),
                ]))]))]),
            'return': [
                'auto bond10\n',
                'iface bond10 inet static\n',
                '    address 10.1.0.14\n',
                '    netmask 255.255.255.0\n',
                '    gateway 10.1.0.1\n',
                '    bond-ad_select 0\n',
                '    bond-downdelay 200\n',
                '    bond-lacp_rate 0\n',
                '    bond-miimon 100\n',
                '    bond-mode 4\n',
                '    bond-slaves eth4 eth5\n',
                '    bond-updelay 0\n',
                '    bond-use_carrier on\n',
                'iface bond10 inet6 static\n',
                '    address 2001:db8:dead:c0::3\n',
                '    netmask 64\n',
                '    gateway 2001:db8:dead:c0::1\n',
                '    bond-ad_select 0\n',
                '    bond-downdelay 200\n',
                '    bond-lacp_rate 0\n',
                '    bond-miimon 100\n',
                '    bond-mode 4\n',
                '    bond-slaves eth4 eth5\n',
                '    bond-updelay 0\n',
                '    bond-use_carrier on\n',
                '\n']},

        # Bond VLAN; with IPv4 address
        {'iface_name': 'bond0.11', 'iface_type': 'vlan', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'ipaddr': '10.7.0.8',
                'netmask': '255.255.255.0',
                'gateway': '10.7.0.1',
                'slaves': 'eth6 eth7',
                'mode': '802.3ad',
                'enable_ipv6': False,
                'noifupdown': True,
                },
            'get_interface': odict([('bond0.11', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('vlan_raw_device', 'bond1'),
                    ('address', '10.7.0.8'),
                    ('netmask', '255.255.255.0'),
                    ('gateway', '10.7.0.1'),
                    ('mode', '802.3ad'),
                    ])),
                ]))]))]),
            'return': [
                'auto bond0.11\n',
                'iface bond0.11 inet static\n',
                '    vlan-raw-device bond1\n',
                '    address 10.7.0.8\n',
                '    netmask 255.255.255.0\n',
                '    gateway 10.7.0.1\n',
                '    mode 802.3ad\n',
                '\n']},

        # Bond; without address
        {'iface_name': 'bond0.12', 'iface_type': 'vlan', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'slaves': 'eth6 eth7',
                'mode': '802.3ad',
                'enable_ipv6': False,
                'noifupdown': True,
                },
            'get_interface': odict([('bond0.12', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('vlan_raw_device', 'bond1'),
                    ('mode', '802.3ad'),
                    ])),
                ]))]))]),
            'return': [
                'auto bond0.12\n',
                'iface bond0.12 inet static\n',
                '    vlan-raw-device bond1\n',
                '    mode 802.3ad\n',
                '\n']},

        # Bridged interface
        {'iface_name': 'br0', 'iface_type': 'bridge', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'ipaddr': '192.168.4.10',
                'netmask': '255.255.255.0',
                'gateway': '192.168.4.1',
                'bridge_ports': 'eth1',
                'enable_ipv6': False,
                'noifupdown': True,
                },
           'get_interface': odict([('br0', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '192.168.4.10'),
                    ('netmask', '255.255.255.0'),
                    ('gateway', '192.168.4.1'),
                    ('bridging', odict([
                        ('ports', 'eth1'),
                    ])),
                    ('bridging_keys', ['ports']),
                    ])),
                ]))]))]),
            'return': [
                'auto br0\n',
                'iface br0 inet static\n',
                '    address 192.168.4.10\n',
                '    netmask 255.255.255.0\n',
                '    gateway 192.168.4.1\n',
                '    bridge_ports eth1\n',
                '\n']},


        # DNS NS as list
        {'iface_name': 'eth13', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'ipaddr': '192.168.4.9',
                'netmask': '255.255.255.0',
                'gateway': '192.168.4.1',
                'enable_ipv6': False,
                'noifupdown': True,
                'dns': ['8.8.8.8', '8.8.4.4'],
                },
            'get_interface': odict([('eth13', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '192.168.4.9'),
                    ('netmask', '255.255.255.0'),
                    ('gateway', '192.168.4.1'),
                    ('dns_nameservers', ['8.8.8.8', '8.8.4.4']),
                    ])),
                ]))]))]),
            'return': [
                'auto eth13\n',
                'iface eth13 inet static\n',
                '    address 192.168.4.9\n',
                '    netmask 255.255.255.0\n',
                '    gateway 192.168.4.1\n',
                '    dns-nameservers 8.8.8.8 8.8.4.4\n',
                '\n']},

        # DNS NS as string
        {'iface_name': 'eth14', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'static',
                'ipaddr': '192.168.4.9',
                'netmask': '255.255.255.0',
                'gateway': '192.168.4.1',
                'enable_ipv6': False,
                'noifupdown': True,
                'dns': '8.8.8.8 8.8.4.4',
                },
            'get_interface': odict([('eth14', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '192.168.4.9'),
                    ('netmask', '255.255.255.0'),
                    ('gateway', '192.168.4.1'),
                    ('dns_nameservers', ['8.8.8.8', '8.8.4.4']),
                    ])),
                ]))]))]),
            'return': [
                'auto eth14\n',
                'iface eth14 inet static\n',
                '    address 192.168.4.9\n',
                '    netmask 255.255.255.0\n',
                '    gateway 192.168.4.1\n',
                '    dns-nameservers 8.8.8.8 8.8.4.4\n',
                '\n']},

        # Loopback; with IPv4 and IPv6 address
        {'iface_name': 'lo15', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'loopback',
                'ipaddr': '192.168.4.9',
                'netmask': '255.255.255.0',
                'gateway': '192.168.4.1',
                'enable_ipv6': True,
                'ipv6proto': 'loopback',
                'ipv6ipaddr': 'fc00::1',
                'ipv6netmask': '128',
                'ipv6_autoconf': False,
                'noifupdown': True,
                },
            'get_interface': odict([('lo15', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'loopback'),
                    ('filename', None),
                    ('address', '192.168.4.9'),
                    ('netmask', '255.255.255.0'),
                    ('gateway', '192.168.4.1'),
                    ])),
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'loopback'),
                    ('filename', None),
                    ('address', 'fc00::1'),
                    ('netmask', 128),
                    ])),
                ]))]))]),
            'return': [
                'auto lo15\n',
                'iface lo15 inet loopback\n',
                '    address 192.168.4.9\n',
                '    netmask 255.255.255.0\n',
                '    gateway 192.168.4.1\n',
                'iface lo15 inet6 loopback\n',
                '    address fc00::1\n',
                '    netmask 128\n',
                '\n']},

        # Loopback; with only IPv6 address; enabled=False
        {'iface_name': 'lo16', 'iface_type': 'eth', 'enabled': False,
            'build_interface': {
                'enable_ipv6': True,
                'ipv6proto': 'loopback',
                'ipv6ipaddr': 'fc00::1',
                'ipv6netmask': '128',
                'ipv6_autoconf': False,
                'noifupdown': True,
                },
            'get_interface': odict([('lo16', odict([('data', odict([
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'loopback'),
                    ('filename', None),
                    ('address', 'fc00::1'),
                    ('netmask', 128),
                    ])),
                ]))]))]),
            'return': [
                'iface lo16 inet6 loopback\n',
                '    address fc00::1\n',
                '    netmask 128\n',
                '\n']},

        # Loopback; without address
        {'iface_name': 'lo17', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'loopback',
                'enable_ipv6': False,
                'noifupdown': True,
                },
            'get_interface': odict([('lo17', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'loopback'),
                    ('filename', None),
                    ])),
                ]))]))]),
            'return': [
                'auto lo17\n',
                'iface lo17 inet loopback\n',
                '\n']},

        # IPv4=DHCP; IPv6=Static; with IPv6 netmask
        {'iface_name': 'eth18', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'dhcp',
                'enable_ipv6': True,
                'ipv6proto': 'static',
                'ipv6ipaddr': '2001:db8:dead:c0::3',
                'ipv6netmask': '64',
                'ipv6gateway': '2001:db8:dead:c0::1',
                'noifupdown': True,
                },
            'get_interface': odict([('eth18', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'dhcp'),
                    ('filename', None),
                    ])),
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '2001:db8:dead:c0::3'),
                    ('netmask', 64),
                    ('gateway', '2001:db8:dead:c0::1'),
                    ])),
                ]))]))]),
            'return': [
                'auto eth18\n',
                'iface eth18 inet dhcp\n',
                'iface eth18 inet6 static\n',
                '    address 2001:db8:dead:c0::3\n',
                '    netmask 64\n',
                '    gateway 2001:db8:dead:c0::1\n',
                '\n']},

        # IPv4=DHCP; IPv6=Static; without IPv6 netmask
        {'iface_name': 'eth19', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'dhcp',
                'enable_ipv6': True,
                'ipv6proto': 'static',
                'ipv6ipaddr': '2001:db8:dead:c0::3/64',
                'ipv6gateway': '2001:db8:dead:c0::1',
                'noifupdown': True,
                },
            'get_interface': odict([('eth19', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'dhcp'),
                    ('filename', None),
                    ])),
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('address', '2001:db8:dead:c0::3/64'),
                    ('gateway', '2001:db8:dead:c0::1'),
                    ])),
                ]))]))]),
            'return': [
                'auto eth19\n',
                'iface eth19 inet dhcp\n',
                'iface eth19 inet6 static\n',
                '    address 2001:db8:dead:c0::3/64\n',
                '    gateway 2001:db8:dead:c0::1\n',
                '\n']},

        # IPv6-only; static with autoconf and accept_ra forced
        {'iface_name': 'eth20', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'ipv6proto': 'static',
                'ipv6ipaddr': '2001:db8:dead:beef::3/64',
                'ipv6gateway': '2001:db8:dead:beef::1',
                'enable_ipv6': True,
                'autoconf': 1,
                'accept_ra': 2,
                'noifupdown': True,
                },
            'get_interface': odict([('eth20', odict([('enabled', True), ('data', odict([
                ('inet6', odict([
                    ('addrfam', 'inet6'),
                    ('proto', 'static'),
                    ('filename', None),
                    ('autoconf', 1),
                    ('address', '2001:db8:dead:beef::3/64'),
                    ('gateway', '2001:db8:dead:beef::1'),
                    ('accept_ra', 2),
                    ])),
                ]))]))]),
            'return': [
                'auto eth20\n',
                'iface eth20 inet6 static\n',
                '    autoconf 1\n',
                '    address 2001:db8:dead:beef::3/64\n',
                '    gateway 2001:db8:dead:beef::1\n',
                '    accept_ra 2\n',
                '\n']},

        # Command hooks
        {'iface_name': 'eth_optmap_test', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'manual',
                'up_cmds': ['echo "interface up"'],
                'down_cmds': ['echo "interface down"'],
                'pre_up_cmds': ['echo "pre-up"'],
                'post_up_cmds': ['echo "post-up"'],
                'pre_down_cmds': ['echo "pre-down"'],
                'post_down_cmds': ['echo "post-down"'],
                },
            'get_interface': odict([('eth_optmap_test', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'manual'),
                    ('filename', None),
                    ('up_cmds', ['echo "interface up"']),
                    ('down_cmds', ['echo "interface down"']),
                    ('pre_up_cmds', ['echo "pre-up"']),
                    ('post_up_cmds', ['echo "post-up"']),
                    ('pre_down_cmds', ['echo "pre-down"']),
                    ('post_down_cmds', ['echo "post-down"']),
                    ])),
                ]))]))]),
            'return': [
                'auto eth_optmap_test\n',
                'iface eth_optmap_test inet manual\n',
                '    up echo "interface up"\n',
                '    down echo "interface down"\n',
                '    pre-up echo "pre-up"\n',
                '    post-up echo "post-up"\n',
                '    pre-down echo "pre-down"\n',
                '    post-down echo "post-down"\n',
                '\n']},

        # hwaddr should map to hwaddress
        {'iface_name': 'eth_hwaddr_test', 'iface_type': 'eth', 'enabled': True,
            'build_interface': {
                'proto': 'manual',
                'hwaddr': '00:11:22:33:44:55',
                },
            'get_interface': odict([('eth_hwaddr_test', odict([('enabled', True), ('data', odict([
                ('inet', odict([
                    ('addrfam', 'inet'),
                    ('proto', 'manual'),
                    ('filename', None),
                    ('hwaddress', '00:11:22:33:44:55'),
                    ])),
                ]))]))]),
            'return': [
                'auto eth_hwaddr_test\n',
                'iface eth_hwaddr_test inet manual\n',
                '    hwaddress 00:11:22:33:44:55\n',
                '\n']},
        ]
# fmt: on


@pytest.fixture
def configure_loader_modules():
    return {debian_ip: {}}


# 'build_bond' function tests: 3


def test_build_bond():
    """
    Test if it create a bond script in /etc/modprobe.d with the passed
    settings and load the bonding kernel module.
    """
    with patch(
        "salt.modules.debian_ip._parse_settings_bond", MagicMock(return_value={})
    ), patch("salt.modules.debian_ip._write_file", MagicMock(return_value=True)):
        mock = MagicMock(return_value=1)
        with patch.dict(debian_ip.__grains__, {"osrelease": mock}):
            mock = MagicMock(return_value=True)
            with patch.dict(
                debian_ip.__salt__, {"kmod.load": mock, "pkg.install": mock}
            ):
                assert debian_ip.build_bond("bond0") == ""


def test_error_message_iface_should_process_non_str_expected():
    values = [1, True, False, "no-kaboom"]
    iface = "ethtest"
    option = "test"
    msg = debian_ip._error_msg_iface(iface, option, values)
    assert msg.endswith("[1|True|False|no-kaboom]"), msg


def test_error_message_network_should_process_non_str_expected():
    values = [1, True, False, "no-kaboom"]
    msg = debian_ip._error_msg_network("fnord", values)
    assert msg.endswith("[1|True|False|no-kaboom]"), msg


def test_build_bond_exception():
    """
    Test if it create a bond script in /etc/modprobe.d with the passed
    settings and load the bonding kernel module.
    """
    with patch(
        "salt.modules.debian_ip._parse_settings_bond", MagicMock(return_value={})
    ):
        mock = MagicMock(return_value=1)
        with patch.dict(debian_ip.__grains__, {"osrelease": mock}):
            mock = MagicMock(side_effect=jinja2.exceptions.TemplateNotFound("error"))
            with patch.object(jinja2.Environment, "get_template", mock):
                assert debian_ip.build_bond("bond0") == ""


def test_build_bond_data():
    """
    Test if it create a bond script in /etc/modprobe.d with the passed
    settings and load the bonding kernel module.
    """
    with patch(
        "salt.modules.debian_ip._parse_settings_bond", MagicMock(return_value={})
    ), patch("salt.modules.debian_ip._read_temp", MagicMock(return_value=True)):
        mock = MagicMock(return_value=1)
        with patch.dict(debian_ip.__grains__, {"osrelease": mock}):
            assert debian_ip.build_bond("bond0", test="True")


# 'build_routes' function tests: 2


def test_build_routes():
    """
    Test if it add route scripts for a network interface using up commands.
    """
    with patch(
        "salt.modules.debian_ip._parse_routes",
        MagicMock(return_value={"routes": []}),
    ), patch(
        "salt.modules.debian_ip._write_file_routes", MagicMock(return_value=True)
    ), patch(
        "salt.modules.debian_ip._read_file", MagicMock(return_value="salt")
    ):
        assert debian_ip.build_routes("eth0") == "saltsalt"


def test_build_routes_exception():
    """
    Test if it add route scripts for a network interface using up commands.
    """
    with patch(
        "salt.modules.debian_ip._parse_routes",
        MagicMock(return_value={"routes": []}),
    ):
        assert debian_ip.build_routes("eth0", test="True")

        mock = MagicMock(side_effect=jinja2.exceptions.TemplateNotFound("err"))
        with patch.object(jinja2.Environment, "get_template", mock):
            assert debian_ip.build_routes("eth0") == ""


# 'down' function tests: 1


def test_down():
    """
    Test if it shutdown a network interface
    """
    assert debian_ip.down("eth0", "slave") is None

    mock = MagicMock(return_value="Salt")
    with patch.dict(debian_ip.__salt__, {"cmd.run": mock}):
        assert debian_ip.down("eth0", "eth") == "Salt"


# 'get_bond' function tests: 1


def test_get_bond():
    """
    Test if it return the content of a bond script
    """
    assert debian_ip.get_bond("bond0") == ""


# '_parse_interfaces' function tests: 1


def test_parse_interfaces(test_interfaces):
    """
    Test if it returns the correct data for parsed configuration file
    """
    with tempfile.NamedTemporaryFile(mode="r", delete=True) as tfile:
        for iface in test_interfaces:
            iname = iface["iface_name"]
            if iface.get("skip_test", False):
                continue
            with salt.utils.files.fopen(str(tfile.name), "w") as fh:
                fh.writelines(iface["return"])
            for inet in ["inet", "inet6"]:
                if inet in iface["get_interface"][iname]["data"]:
                    iface["get_interface"][iname]["data"][inet]["filename"] = str(
                        tfile.name
                    )
            assert (
                debian_ip._parse_interfaces([str(tfile.name)]) == iface["get_interface"]
            )


# 'get_interface' function tests: 1


def test_get_interface(test_interfaces):
    """
    Test if it return the contents of an interface script
    """
    for iface in test_interfaces:
        if iface.get("skip_test", False):
            continue
        with patch.object(
            debian_ip,
            "_parse_interfaces",
            MagicMock(return_value=iface["get_interface"]),
        ):
            assert debian_ip.get_interface(iface["iface_name"]) == iface["return"]


# 'build_interface' function tests: 1


def test_build_interface(test_interfaces):
    """
    Test if it builds an interface script for a network interface.
    """
    with patch(
        "salt.modules.debian_ip._write_file_ifaces", MagicMock(return_value="salt")
    ):
        assert debian_ip.build_interface("eth0", "eth", "enabled") == [
            "s\n",
            "a\n",
            "l\n",
            "t\n",
        ]

        assert debian_ip.build_interface("eth0", "eth", "enabled", test="True")

        with patch.object(
            debian_ip, "_parse_settings_eth", MagicMock(return_value={"routes": []})
        ):
            for eth_t in ["bridge", "slave", "bond"]:
                pytest.raises(
                    AttributeError,
                    debian_ip.build_interface,
                    "eth0",
                    eth_t,
                    "enabled",
                )

        assert debian_ip.build_interface("eth0", "eth", "enabled", test="True")

    with tempfile.NamedTemporaryFile(mode="r", delete=True) as tfile:
        with patch("salt.modules.debian_ip._DEB_NETWORK_FILE", str(tfile.name)):
            for iface in test_interfaces:
                if iface.get("skip_test", False):
                    continue
                # Skip tests that require __salt__['pkg.install']()
                if iface["iface_type"] in ["bridge", "pppoe", "vlan"]:
                    continue
                assert (
                    debian_ip.build_interface(
                        iface=iface["iface_name"],
                        iface_type=iface["iface_type"],
                        enabled=iface["enabled"],
                        interface_file=tfile.name,
                        **iface["build_interface"]
                    )
                    == iface["return"]
                )


# 'up' function tests: 1


def test_up():
    """
    Test if it start up a network interface
    """
    assert debian_ip.down("eth0", "slave") is None

    mock = MagicMock(return_value="Salt")
    with patch.dict(debian_ip.__salt__, {"cmd.run": mock}):
        assert debian_ip.up("eth0", "eth") == "Salt"


# 'get_network_settings' function tests: 1


def test_get_network_settings():
    """
    Test if it return the contents of the global network script.
    """
    with patch.dict(
        debian_ip.__grains__, {"osfullname": "Ubuntu", "osrelease": "14"}
    ), patch(
        "salt.modules.debian_ip._parse_hostname",
        MagicMock(return_value="SaltStack"),
    ), patch(
        "salt.modules.debian_ip._parse_domainname",
        MagicMock(return_value="saltstack.com"),
    ):
        mock_avai = MagicMock(return_value=True)
        with patch.dict(
            debian_ip.__salt__,
            {"service.available": mock_avai, "service.status": mock_avai},
        ):
            assert debian_ip.get_network_settings() == [
                "NETWORKING=yes\n",
                "HOSTNAME=SaltStack\n",
                "DOMAIN=saltstack.com\n",
            ]

            mock = MagicMock(side_effect=jinja2.exceptions.TemplateNotFound("error"))
            with patch.object(jinja2.Environment, "get_template", mock):
                assert debian_ip.get_network_settings() == ""


# 'get_routes' function tests: 1


def test_get_routes():
    """
    Test if it return the routes for the interface
    """
    with patch("salt.modules.debian_ip._read_file", MagicMock(return_value="salt")):
        assert debian_ip.get_routes("eth0") == "saltsalt"


# 'apply_network_settings' function tests: 1


@pytest.mark.slow_test
def test_apply_network_settings():
    """
    Test if it apply global network configuration.
    """
    mock = MagicMock(return_value=True)
    with patch.dict(
        debian_ip.__salt__,
        {"network.mod_hostname": mock, "service.stop": mock, "service.start": mock},
    ):
        assert debian_ip.apply_network_settings() is True


# 'build_network_settings' function tests: 1


def test_build_network_settings():
    """
    Test if it build the global network script.
    """
    with patch(
        "salt.modules.debian_ip._parse_network_settings",
        MagicMock(
            return_value={
                "networking": "yes",
                "hostname": "Salt.saltstack.com",
                "domainname": "saltstack.com",
                "search": "test.saltstack.com",
            }
        ),
    ), patch(
        "salt.modules.debian_ip._write_file_network", MagicMock(return_value=True)
    ):
        with patch.dict(
            debian_ip.__grains__, {"osfullname": "Ubuntu", "osrelease": "14"}
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(
                debian_ip.__salt__,
                {
                    "service.available": mock,
                    "service.disable": mock,
                    "service.enable": mock,
                },
            ):
                assert debian_ip.build_network_settings() == [
                    "NETWORKING=yes\n",
                    "HOSTNAME=Salt\n",
                    "DOMAIN=saltstack.com\n",
                    "SEARCH=test.saltstack.com\n",
                ]

                mock = MagicMock(
                    side_effect=jinja2.exceptions.TemplateNotFound("error")
                )
                with patch.object(jinja2.Environment, "get_template", mock):
                    assert debian_ip.build_network_settings() == ""

        with patch.dict(
            debian_ip.__grains__, {"osfullname": "Ubuntu", "osrelease": "10"}
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(
                debian_ip.__salt__,
                {
                    "service.available": mock,
                    "service.disable": mock,
                    "service.enable": mock,
                },
            ):
                mock = MagicMock(
                    side_effect=jinja2.exceptions.TemplateNotFound("error")
                )
                with patch.object(jinja2.Environment, "get_template", mock):
                    assert debian_ip.build_network_settings() == ""

                with patch.object(
                    debian_ip, "_read_temp", MagicMock(return_value=True)
                ):
                    assert debian_ip.build_network_settings(test="True")


def test_when_no_adapters_are_passed_to_filter_none_should_be_returned():
    no_adapters = {}

    actual_adapters = debian_ip._filter_malformed_interfaces(adapters=no_adapters)

    assert not actual_adapters


def test_when_adapters_only_contains_source_then_source_should_be_returned():
    souce_only_adapters = {"source": "fnord"}
    expected_adapters = souce_only_adapters.copy()

    actual_adapters = debian_ip._filter_malformed_interfaces(
        adapters=souce_only_adapters
    )

    assert actual_adapters == expected_adapters


def test_when_data_is_missing_from_adapters_they_should_not_be_in_result():
    missing_data_adapters = {"no data adapter": "no dayta is here"}
    expected_adapters = {}
    actual_adapters = debian_ip._filter_malformed_interfaces(
        adapters=missing_data_adapters
    )

    assert actual_adapters == expected_adapters


def test_when_data_in_adapters_and_no_inet_or_inet6_in_data_segment_then_original_data_should_be_returned():
    expected_adapters = {
        "some cool adapter": {"data": {}},
        "some other adapter": {"data": {}},
        "yet another neat adapter": {"data": {}},
    }

    no_inet_data_adapters = {"no data adapter": "this one should be gone"}
    no_inet_data_adapters.update(expected_adapters)

    actual_adapters = debian_ip._filter_malformed_interfaces(
        adapters=no_inet_data_adapters
    )
    assert actual_adapters == expected_adapters


def test_when_opts_are_in_data_sorted_opt_keys_should_be_added():
    comprehensive_adapters = {
        "source": "keep me here",
        "no data adapter": "lulz",
        "adapter 1": {
            "data": {
                "inet": {
                    "ethtool": {
                        "5": {},
                        "4": {},
                        "2": {},
                        "3": {},
                        "1": {},
                    },
                    "bonding": {
                        "4": {},
                        "5": {},
                        "3": {},
                        "2": {},
                        "1": {},
                    },
                    "bridging": {
                        "1": {},
                        "5": {},
                        "4": {},
                        "2": {},
                        "3": {},
                    },
                },
            },
        },
    }

    expected_adapters = {
        "source": "keep me here",
        "adapter 1": {
            "data": {
                "inet": {
                    "ethtool_keys": ["1", "2", "3", "4", "5"],
                    "bonding_keys": ["1", "2", "3", "4", "5"],
                    "bridging_keys": ["1", "2", "3", "4", "5"],
                    "ethtool": {
                        "5": {},
                        "4": {},
                        "3": {},
                        "2": {},
                        "1": {},
                    },
                    "bonding": {
                        "5": {},
                        "4": {},
                        "3": {},
                        "2": {},
                        "1": {},
                    },
                    "bridging": {
                        "5": {},
                        "4": {},
                        "3": {},
                        "2": {},
                        "1": {},
                    },
                },
            },
        },
    }
    actual_adapters = debian_ip._filter_malformed_interfaces(
        adapters=comprehensive_adapters
    )
    assert actual_adapters == expected_adapters



import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


def test_saltcheck_run_test(salt_ssh_cli):
    """
    test saltcheck.run_test with salt-ssh
    """
    saltcheck_test = {
        "module_and_function": "test.echo",
        "assertion": "assertEqual",
        "expected-return": "Test Works",
        "args": ["Test Works"],
    }
    ret = salt_ssh_cli.run("saltcheck.run_test", test=saltcheck_test)
    assert ret.returncode == 0
    assert ret.data
    assert "status" in ret.data
    assert ret.data["status"] == "Pass"


@pytest.mark.skip_on_aarch64
def test_saltcheck_state(salt_ssh_cli):
    """
    saltcheck.run_state_tests
    """
    ret = salt_ssh_cli.run("saltcheck.run_state_tests", "validate-saltcheck")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data[0]
    assert "validate-saltcheck" in ret.data[0]
    state_result = ret.data[0]["validate-saltcheck"]
    assert "echo_test_hello" in state_result
    assert "status" in state_result["echo_test_hello"]
    assert state_result["echo_test_hello"]["status"] == "Pass"



import os
import time

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def file(modules):
    return modules.file


@pytest.fixture(scope="function")
def single_empty_dir(tmp_path):
    yield str(tmp_path)


@pytest.fixture(scope="function")
def single_dir_with_file(tmp_path):
    file = tmp_path / "stuff.txt"
    file.write_text("things")
    yield str(tmp_path)


@pytest.fixture(scope="function")
def nested_empty_dirs(tmp_path):
    num_root = 2
    num_mid = 4
    num_last = 2
    for root in range(1, num_root + 1):
        for mid in range(1, num_mid + 1):
            for last in range(1, num_last + 1):
                nest = tmp_path / f"root{root}" / f"mid{mid}" / f"last{last}"
                nest.mkdir(parents=True, exist_ok=True)
                if last % 2:
                    now = time.time()
                    old = now - (2 * 86400)
                    os.utime(str(nest), (old, old))
    yield str(tmp_path)


@pytest.fixture(scope="function")
def nested_dirs_with_files(tmp_path):
    num_root = 2
    num_mid = 4
    num_last = 2
    for root in range(1, num_root + 1):
        for mid in range(1, num_mid + 1):
            for last in range(1, num_last + 1):
                nest = tmp_path / f"root{root}" / f"mid{mid}" / f"last{last}"
                nest.mkdir(parents=True, exist_ok=True)
                if last % 2:
                    last_file = nest / "stuff.txt"
                    last_file.write_text("things")
    yield str(tmp_path)


def test_rmdir_success_with_default_options(file, single_empty_dir):
    assert file.rmdir(single_empty_dir) is True
    assert not os.path.isdir(single_empty_dir)
    assert not os.path.exists(single_empty_dir)


def test_rmdir_failure_with_default_options(file, single_dir_with_file):
    assert file.rmdir(single_dir_with_file) is False
    assert os.path.isdir(single_dir_with_file)


def test_rmdir_single_dir_success_with_recurse(file, single_empty_dir):
    assert file.rmdir(single_empty_dir, recurse=True) is True
    assert not os.path.isdir(single_empty_dir)
    assert not os.path.exists(single_empty_dir)


def test_rmdir_single_dir_failure_with_recurse(file, single_dir_with_file):
    assert file.rmdir(single_dir_with_file, recurse=True) is False
    assert os.path.isdir(single_dir_with_file)


def test_rmdir_nested_empty_dirs_failure_with_default_options(file, nested_empty_dirs):
    assert file.rmdir(nested_empty_dirs) is False
    assert os.path.isdir(nested_empty_dirs)


def test_rmdir_nested_empty_dirs_success_with_recurse(file, nested_empty_dirs):
    assert file.rmdir(nested_empty_dirs, recurse=True) is True
    assert not os.path.isdir(nested_empty_dirs)
    assert not os.path.exists(nested_empty_dirs)


def test_rmdir_nested_dirs_with_files_failure_with_recurse(
    file, nested_dirs_with_files
):
    assert file.rmdir(nested_dirs_with_files, recurse=True) is False
    assert os.path.isdir(nested_dirs_with_files)


def test_rmdir_verbose_nested_dirs_with_files_failure_with_recurse(
    file, nested_dirs_with_files
):
    ret = file.rmdir(nested_dirs_with_files, recurse=True, verbose=True)
    assert ret["result"] is False
    assert len(ret["deleted"]) == 8
    assert len(ret["errors"]) == 19
    assert os.path.isdir(nested_dirs_with_files)


def test_rmdir_verbose_success(file, single_empty_dir):
    ret = file.rmdir(single_empty_dir, verbose=True)
    assert ret["result"] is True
    assert ret["deleted"][0] == single_empty_dir
    assert not ret["errors"]
    assert not os.path.isdir(single_empty_dir)
    assert not os.path.exists(single_empty_dir)


def test_rmdir_verbose_failure(file, single_dir_with_file):
    ret = file.rmdir(single_dir_with_file, verbose=True)
    assert ret["result"] is False
    assert not ret["deleted"]
    assert ret["errors"][0][0] == single_dir_with_file
    assert os.path.isdir(single_dir_with_file)


def test_rmdir_nested_empty_dirs_recurse_older_than(file, nested_empty_dirs):
    ret = file.rmdir(nested_empty_dirs, recurse=True, verbose=True, older_than=1)
    assert ret["result"] is True
    assert len(ret["deleted"]) == 8
    assert len(ret["errors"]) == 0
    assert os.path.isdir(nested_empty_dirs)


def test_rmdir_nested_empty_dirs_recurse_not_older_than(file, nested_empty_dirs):
    ret = file.rmdir(nested_empty_dirs, recurse=True, verbose=True, older_than=3)
    assert ret["result"] is True
    assert len(ret["deleted"]) == 0
    assert len(ret["errors"]) == 0
    assert os.path.isdir(nested_empty_dirs)



import os
import pathlib
import subprocess

import packaging.version
import psutil
import pytest

pytestmark = [
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture(autouse=True)
def _skip_on_less_than_3006_1(install_salt):
    if packaging.version.parse(install_salt.version) <= packaging.version.parse(
        "3006.1"
    ):
        pytest.skip(
            "Multi-minion script only available on versions greater than 3006.1"
        )


@pytest.fixture
def mm_script(install_salt):
    yield install_salt.ssm_bin.parent / "multi-minion.ps1"


@pytest.fixture(scope="function")
def mm_conf(mm_script):
    yield pathlib.Path(os.getenv("LocalAppData"), "Salt Project", "Salt", "conf")
    subprocess.run(
        ["powershell", str(mm_script).replace(" ", "' '"), "-d"],
        capture_output=True,
        check=False,
        text=True,
    )


def test_script_present(mm_script):
    """
    Ensure the multi-minion.ps1 file is present in the root of the installation
    """
    assert mm_script.exists()


def test_install(mm_script, mm_conf):
    """
    Install a second minion with default settings. Should create a minion config
    file in Local AppData
    """
    ret = subprocess.run(
        ["powershell", str(mm_script).replace(" ", "' '")],
        capture_output=True,
        check=False,
        text=True,
    )
    assert ret.returncode == 0, ret.stderr
    conf_file = mm_conf / "minion"
    assert conf_file.exists()
    assert conf_file.read_text().find("master: salt") > -1


def test_install_master(mm_script, mm_conf):
    """
    Install a second minion and set the master to spongebob
    """
    ret = subprocess.run(
        ["powershell", str(mm_script).replace(" ", "' '"), "-m", "spongebob"],
        capture_output=True,
        check=False,
        text=True,
    )
    assert ret.returncode == 0, ret.stderr
    conf_file = mm_conf / "minion"
    assert conf_file.exists()
    assert conf_file.read_text().find("master: spongebob") > -1


def test_install_prefix(mm_script, mm_conf):
    """
    Install a second minion and add a prefix to the minion id
    """
    ret = subprocess.run(
        ["powershell", str(mm_script).replace(" ", "' '"), "-p", "squarepants"],
        capture_output=True,
        check=False,
        text=True,
    )
    assert ret.returncode == 0, ret.stderr
    conf_file = mm_conf / "minion"
    assert conf_file.exists()
    assert conf_file.read_text().find("id: squarepants") > -1


def test_install_log_level(mm_script, mm_conf):
    """
    Install a second minion and set the log level in the log file to debug
    """
    ret = subprocess.run(
        ["powershell", str(mm_script).replace(" ", "' '"), "-l", "debug"],
        capture_output=True,
        check=False,
        text=True,
    )
    assert ret.returncode == 0, ret.stderr
    conf_file = mm_conf / "minion"
    assert conf_file.exists()
    assert conf_file.read_text().find("log_level_logfile: debug") > -1


def test_install_start(mm_script, mm_conf):
    """
    Install a second minion and start that minion in a hidden process
    """
    ret = subprocess.run(
        ["powershell", str(mm_script).replace(" ", "' '"), "-s"],
        capture_output=True,
        check=False,
        text=True,
    )
    assert ret.returncode == 0, ret.stderr
    conf_file = mm_conf / "minion"
    assert conf_file.exists()
    assert conf_file.read_text().find("master: salt") > -1

    found = False
    for p in psutil.process_iter(["cmdline", "name"]):
        if p.info["name"] and p.info["name"] == "salt-minion.exe":
            if f"{mm_conf}" in p.info["cmdline"]:
                found = True
    assert found is True



"""
    saltrepo
    ~~~~~~~~

    SaltStack Repository Sphinx directives
"""


def source_read_handler(app, docname, source):
    if "|repo_primary_branch|" in source[0]:
        source[0] = source[0].replace(
            "|repo_primary_branch|", app.config.html_context["repo_primary_branch"]
        )


def setup(app):
    app.connect("source-read", source_read_handler)

    return {
        "version": "builtin",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import pytest

import salt.utils.json
from salt.modules import http
from salt.pillar import http_json


@pytest.fixture
def configure_loader_modules():
    return {
        http_json: {
            "__salt__": {
                "http.query": http.query,
            },
        },
        http: {
            "__opts__": {},
        },
    }


@pytest.mark.requires_network
@pytest.mark.parametrize("backend", ["requests", "tornado", "urllib2"])
def test_ext_pillar_can_take_http_query_kwargs(backend, httpserver):
    response = {
        "dict": {
            "backend": backend,
            "pillar_type": "http_json",
        },
    }
    header_dict = {"custom-backend-header": backend}

    # If the headers in header_dict are not in the request, httpserver will return an empty dictionary, so we know it will fail
    httpserver.expect_request(
        f"/http_json_pillar/{backend}",
        headers={"custom-backend-header": backend},
    ).respond_with_data(salt.utils.json.dumps(response), content_type="text/plain")
    url = httpserver.url_for(f"/http_json_pillar/{backend}")

    actual = http_json.ext_pillar("test-minion-id", {}, url, header_dict=header_dict)
    assert actual == response


@pytest.mark.requires_network
@pytest.mark.parametrize("backend", ["requests", "tornado", "urllib2"])
def test_ext_pillar_namespace(backend, httpserver):
    response = {
        "dict": {
            "backend": backend,
            "pillar_type": "http_json",
        },
    }
    header_dict = {"custom-backend-header": backend}
    namespace = "test_namespace"

    # If the headers in header_dict are not in the request, httpserver will return an empty dictionary, so we know it will fail
    httpserver.expect_request(
        f"/http_json_pillar/{backend}",
        headers={"custom-backend-header": backend},
    ).respond_with_data(salt.utils.json.dumps(response), content_type="text/plain")
    url = httpserver.url_for(f"/http_json_pillar/{backend}")

    actual = http_json.ext_pillar(
        "test-minion-id", {}, url, header_dict=header_dict, namespace=namespace
    )
    assert actual == {namespace: response}



"""
These only test the provider selection and verification logic, they do not init
any remotes.
"""

import logging
import os
import pathlib
import signal
import time

import pytest
from saltfactories.utils import random_string

import salt.ext.tornado.ioloop
import salt.fileserver.gitfs
import salt.utils.files
import salt.utils.gitfs
import salt.utils.path
import salt.utils.platform
import salt.utils.process
from salt.utils.immutabletypes import freeze
from salt.utils.platform import get_machine_identifier as _get_machine_identifier
from salt.utils.verify import verify_env

try:
    import pwd
except ImportError:
    import salt.utils.win_functions

pytestmark = [pytest.mark.skip_on_windows]

log = logging.getLogger(__name__)


def _get_user():
    """
    Get the user associated with the current process.
    """
    if salt.utils.platform.is_windows():
        return salt.utils.win_functions.get_current_user(with_domain=False)
    return pwd.getpwuid(os.getuid())[0]


def _clear_instance_map():
    try:
        del salt.utils.gitfs.GitFS.instance_map[
            salt.ext.tornado.ioloop.IOLoop.current()
        ]
    except KeyError:
        pass


class MyMockedGitProvider:
    """
    mocked GitFS provider leveraging tmp_path
    """

    def __init__(
        self,
        salt_factories_default_root_dir,
        temp_salt_master,
        temp_salt_minion,
        tmp_path,
    ):
        self._tmp_name = str(tmp_path)

        self._root_dir = str(salt_factories_default_root_dir)
        self._master_cfg = str(temp_salt_master.config["conf_file"])
        self._minion_cfg = str(temp_salt_minion.config["conf_file"])
        self._user = _get_user()

        tmp_name = self._tmp_name.join("/git_test")
        pathlib.Path(tmp_name).mkdir(exist_ok=True, parents=True)

        class MockedProvider(
            salt.utils.gitfs.GitProvider
        ):  # pylint: disable=abstract-method
            def __init__(
                self,
                opts,
                remote,
                per_remote_defaults,
                per_remote_only,
                override_params,
                cache_root,
                role="gitfs",
            ):
                self.provider = "mocked"
                self.fetched = False
                super().__init__(
                    opts,
                    remote,
                    per_remote_defaults,
                    per_remote_only,
                    override_params,
                    cache_root,
                    role,
                )

            def init_remote(self):
                self.gitdir = salt.utils.path.join(tmp_name, ".git")
                self.repo = True
                new = False
                return new

            def envs(self):
                return ["base"]

            def _fetch(self):
                self.fetched = True

        # Clear the instance map so that we make sure to create a new instance
        # for this test class.
        _clear_instance_map()

        git_providers = {
            "mocked": MockedProvider,
        }
        gitfs_remotes = ["file://repo1.git", {"file://repo2.git": [{"name": "repo2"}]}]

        self.opts = self.get_temp_config(
            "master",
            gitfs_remotes=gitfs_remotes,
            verified_gitfs_provider="mocked",
        )
        self.main_class = salt.utils.gitfs.GitFS(
            self.opts,
            self.opts["gitfs_remotes"],
            per_remote_overrides=salt.fileserver.gitfs.PER_REMOTE_OVERRIDES,
            per_remote_only=salt.fileserver.gitfs.PER_REMOTE_ONLY,
            git_providers=git_providers,
        )

    def cleanup(self):
        # Providers are preserved with GitFS's instance_map
        for remote in self.main_class.remotes:
            remote.fetched = False
        del self.main_class

    def get_temp_config(self, config_for, **config_overrides):

        rootdir = config_overrides.get("root_dir", self._root_dir)

        if not pathlib.Path(rootdir).exists():
            pathlib.Path(rootdir).mkdir(exist_ok=True, parents=True)

        conf_dir = config_overrides.pop(
            "conf_dir", str(pathlib.PurePath(rootdir).joinpath("conf"))
        )

        for key in ("cachedir", "pki_dir", "sock_dir"):
            if key not in config_overrides:
                config_overrides[key] = key
        if "log_file" not in config_overrides:
            config_overrides["log_file"] = f"logs/{config_for}.log".format()
        if "user" not in config_overrides:
            config_overrides["user"] = self._user
        config_overrides["root_dir"] = rootdir

        cdict = self.get_config(
            config_for,
            from_scratch=True,
        )

        if config_for in ("master", "client_config"):
            rdict = salt.config.apply_master_config(config_overrides, cdict)
        if config_for == "minion":
            minion_id = (
                config_overrides.get("id")
                or config_overrides.get("minion_id")
                or cdict.get("id")
                or cdict.get("minion_id")
                or random_string("temp-minion-")
            )
            config_overrides["minion_id"] = config_overrides["id"] = minion_id
            rdict = salt.config.apply_minion_config(
                config_overrides, cdict, cache_minion_id=False, minion_id=minion_id
            )

        verify_env(
            [
                pathlib.PurePath(rdict["pki_dir"]).joinpath("minions"),
                pathlib.PurePath(rdict["pki_dir"]).joinpath("minions_pre"),
                pathlib.PurePath(rdict["pki_dir"]).joinpath("minions_rejected"),
                pathlib.PurePath(rdict["pki_dir"]).joinpath("minions_denied"),
                pathlib.PurePath(rdict["cachedir"]).joinpath("jobs"),
                pathlib.PurePath(rdict["cachedir"]).joinpath("tokens"),
                pathlib.PurePath(rdict["root_dir"]).joinpath("cache", "tokens"),
                pathlib.PurePath(rdict["pki_dir"]).joinpath("accepted"),
                pathlib.PurePath(rdict["pki_dir"]).joinpath("rejected"),
                pathlib.PurePath(rdict["pki_dir"]).joinpath("pending"),
                pathlib.PurePath(rdict["log_file"]).parent,
                rdict["sock_dir"],
                conf_dir,
            ],
            self._user,
            root_dir=rdict["root_dir"],
        )

        rdict["conf_file"] = pathlib.PurePath(conf_dir).joinpath(config_for)
        with salt.utils.files.fopen(rdict["conf_file"], "w") as wfh:
            salt.utils.yaml.safe_dump(rdict, wfh, default_flow_style=False)
        return rdict

    def get_config(
        self,
        config_for,
        from_scratch=False,
    ):
        if from_scratch:
            if config_for in ("master"):
                return salt.config.master_config(self._master_cfg)
            elif config_for in ("minion"):
                return salt.config.minion_config(self._minion_cfg)
            elif config_for == "client_config":
                return salt.config_client_config(self._master_cfg)
        if config_for not in ("master", "minion", "client_config"):
            if config_for in ("master"):
                return freeze(salt.config.master_config(self._master_cfg))
            elif config_for in ("minion"):
                return freeze(salt.config.minion_config(self._minion_cfg))
            elif config_for == "client_config":
                return freeze(salt.config.client_config(self._master_cfg))

        log.error(
            "Should not reach this section of code for get_config, missing support for input config_for %s",
            config_for,
        )

        # at least return master's config
        return freeze(salt.config.master_config(self._master_cfg))

    @property
    def config_dir(self):
        return str(pathlib.PurePath(self._master_cfg).parent)

    def get_config_dir(self):
        log.warning("Use the config_dir attribute instead of calling get_config_dir()")
        return self.config_dir

    def get_config_file_path(self, filename):
        if filename == "master":
            return str(self._master_cfg)

        if filename == "minion":
            return str(self._minion_cfg)

        return str(self._master_cfg)

    @property
    def master_opts(self):
        """
        Return the options used for the master
        """
        return self.get_config("master")

    @property
    def minion_opts(self):
        """
        Return the options used for the minion
        """
        return self.get_config("minion")


@pytest.fixture
def main_class(
    salt_factories_default_root_dir,
    temp_salt_master,
    temp_salt_minion,
    tmp_path,
):
    my_git_base = MyMockedGitProvider(
        salt_factories_default_root_dir,
        temp_salt_master,
        temp_salt_minion,
        tmp_path,
    )
    yield my_git_base.main_class

    my_git_base.cleanup()


def test_update_all(main_class):
    main_class.update()
    assert len(main_class.remotes) == 2, "Wrong number of remotes"
    assert main_class.remotes[0].fetched
    assert main_class.remotes[1].fetched


def test_update_by_name(main_class):
    main_class.update("repo2")
    assert len(main_class.remotes) == 2, "Wrong number of remotes"
    assert not main_class.remotes[0].fetched
    assert main_class.remotes[1].fetched


def test_update_by_id_and_name(main_class):
    main_class.update([("file://repo1.git", None)])
    assert len(main_class.remotes) == 2, "Wrong number of remotes"
    assert main_class.remotes[0].fetched
    assert not main_class.remotes[1].fetched


def test_get_cachedir_basename(main_class):
    assert main_class.remotes[0].get_cache_basename() == "_"
    assert main_class.remotes[1].get_cache_basename() == "_"


def test_git_provider_mp_lock_and_clear_lock(main_class):
    """
    Check that lock is released after provider.lock()
    and that lock is released after provider.clear_lock()
    """
    provider = main_class.remotes[0]
    provider.lock()
    # check that lock has been released
    assert provider._master_lock.acquire(timeout=5)
    provider._master_lock.release()

    provider.clear_lock()
    # check that lock has been released
    assert provider._master_lock.acquire(timeout=5)
    provider._master_lock.release()


@pytest.mark.slow_test
@pytest.mark.timeout_unless_on_windows(120)
def test_git_provider_mp_lock_timeout(main_class):
    """
    Check that lock will time out if master lock is locked.
    """
    provider = main_class.remotes[0]
    # Hijack the lock so git provider is fooled into thinking another instance is doing somthing.
    assert provider._master_lock.acquire(timeout=5)
    try:
        # git provider should raise timeout error to avoid lock race conditions
        pytest.raises(TimeoutError, provider.lock)
    finally:
        provider._master_lock.release()


@pytest.mark.slow_test
@pytest.mark.timeout_unless_on_windows(120)
def test_git_provider_mp_clear_lock_timeout(main_class):
    """
    Check that clear lock will time out if master lock is locked.
    """
    provider = main_class.remotes[0]
    # Hijack the lock so git provider is fooled into thinking another instance is doing somthing.
    assert provider._master_lock.acquire(timeout=5)
    try:
        # git provider should raise timeout error to avoid lock race conditions
        pytest.raises(TimeoutError, provider.clear_lock)
    finally:
        provider._master_lock.release()


@pytest.mark.slow_test
@pytest.mark.timeout_unless_on_windows(120)
def test_git_provider_mp_gen_lock(main_class, caplog):
    """
    Check that gen_lock is obtains lock, and then releases, provider.lock()
    """
    # get machine_identifier
    mach_id = _get_machine_identifier().get("machine_id", "no_machine_id_available")
    cur_pid = os.getpid()

    test_msg1 = (
        f"Set update lock for gitfs remote 'file://repo1.git' on machine_id '{mach_id}'"
    )
    test_msg2 = (
        "Attempting to remove 'update' lock for 'gitfs' remote 'file://repo1.git' "
        "due to lock_set1 'True' or lock_set2"
    )
    test_msg3 = f"Removed update lock for gitfs remote 'file://repo1.git' on machine_id '{mach_id}'"

    provider = main_class.remotes[0]

    # loop seeing if the test can be made to mess up a lock/unlock sequence
    max_count = 10000
    count = 0
    while count < max_count:
        count = count + 1
        caplog.clear()
        with caplog.at_level(logging.DEBUG):
            provider.fetch()

        assert test_msg1 in caplog.text
        assert test_msg2 in caplog.text
        assert test_msg3 in caplog.text

    caplog.clear()


@pytest.mark.slow_test
@pytest.mark.timeout_unless_on_windows(120)
def test_git_provider_mp_lock_dead_pid(main_class, caplog):
    """
    Check that lock obtains lock, if previous pid in lock file doesn't exist for same machine id
    """
    # get machine_identifier
    mach_id = _get_machine_identifier().get("machine_id", "no_machine_id_available")
    cur_pid = os.getpid()

    test_msg1 = (
        f"Set update lock for gitfs remote 'file://repo1.git' on machine_id '{mach_id}'"
    )
    test_msg3 = f"Removed update lock for gitfs remote 'file://repo1.git' on machine_id '{mach_id}'"

    provider = main_class.remotes[0]
    provider.lock()
    # check that lock has been released
    assert provider._master_lock.acquire(timeout=5)

    # get lock file and manipulate it for a dead pid
    file_name = provider._get_lock_file("update")
    dead_pid = 1234  # give it non-existant pid
    test_msg2 = (
        f"gitfs_global_lock is enabled and update lockfile {file_name} "
        "is present for gitfs remote 'file://repo1.git' on machine_id "
        f"{mach_id} with pid '{dead_pid}'. Process {dead_pid} obtained "
        f"the lock for machine_id {mach_id}, current machine_id {mach_id} "
        "but this process is not running. The update may have been "
        "interrupted.  Given this process is for the same machine the "
        "lock will be reallocated to new process"
    )

    # remove existing lock file and write fake lock file with bad pid
    assert pathlib.Path(file_name).is_file()
    pathlib.Path(file_name).unlink()

    try:
        # write lock file similar to salt/utils/gitfs.py
        fh_ = os.open(file_name, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        with os.fdopen(fh_, "wb"):
            # Write the lock file and close the filehandle
            os.write(fh_, salt.utils.stringutils.to_bytes(str(dead_pid)))
            os.write(fh_, salt.utils.stringutils.to_bytes("\n"))
            os.write(fh_, salt.utils.stringutils.to_bytes(str(mach_id)))
            os.write(fh_, salt.utils.stringutils.to_bytes("\n"))

    except OSError as exc:
        log.error(
            "Failed to write fake dead pid lock file %s, exception %s", file_name, exc
        )

    finally:
        provider._master_lock.release()

    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        provider.lock()
        # check that lock has been released
        assert provider._master_lock.acquire(timeout=5)
        provider._master_lock.release()

        provider.clear_lock()
        # check that lock has been released
        assert provider._master_lock.acquire(timeout=5)
        provider._master_lock.release()

    assert test_msg1 in caplog.text
    assert test_msg2 in caplog.text
    assert test_msg3 in caplog.text
    caplog.clear()


@pytest.mark.slow_test
@pytest.mark.timeout_unless_on_windows(120)
def test_git_provider_mp_lock_bad_machine(main_class, caplog):
    """
    Check that lock obtains lock, if previous pid in lock file doesn't exist for same machine id
    """
    # get machine_identifier
    mach_id = _get_machine_identifier().get("machine_id", "no_machine_id_available")
    cur_pid = os.getpid()

    provider = main_class.remotes[0]
    provider.lock()
    # check that lock has been released
    assert provider._master_lock.acquire(timeout=5)

    # get lock file and manipulate it for a dead pid
    file_name = provider._get_lock_file("update")
    bad_mach_id = "abcedf0123456789"  # give it non-existant pid

    test_msg1 = (
        f"gitfs_global_lock is enabled and update lockfile {file_name} "
        "is present for gitfs remote 'file://repo1.git' on machine_id "
        f"{mach_id} with pid '{cur_pid}'. Process {cur_pid} obtained "
        f"the lock for machine_id {bad_mach_id}, current machine_id {mach_id}"
    )
    test_msg2 = f"Removed update lock for gitfs remote 'file://repo1.git' on machine_id '{mach_id}'"

    # remove existing lock file and write fake lock file with bad pid
    assert pathlib.Path(file_name).is_file()
    pathlib.Path(file_name).unlink()

    try:
        # write lock file similar to salt/utils/gitfs.py
        fh_ = os.open(file_name, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        with os.fdopen(fh_, "wb"):
            # Write the lock file and close the filehandle
            os.write(fh_, salt.utils.stringutils.to_bytes(str(cur_pid)))
            os.write(fh_, salt.utils.stringutils.to_bytes("\n"))
            os.write(fh_, salt.utils.stringutils.to_bytes(str(bad_mach_id)))
            os.write(fh_, salt.utils.stringutils.to_bytes("\n"))

    except OSError as exc:
        log.error(
            "Failed to write fake dead pid lock file %s, exception %s", file_name, exc
        )

    finally:
        provider._master_lock.release()

    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        provider.lock()
        # check that lock has been released
        assert provider._master_lock.acquire(timeout=5)
        provider._master_lock.release()

        provider.clear_lock()
        # check that lock has been released
        assert provider._master_lock.acquire(timeout=5)
        provider._master_lock.release()

    assert test_msg1 in caplog.text
    assert test_msg2 in caplog.text
    caplog.clear()


class KillProcessTest(salt.utils.process.SignalHandlingProcess):
    """
    Test process for which to kill and check lock resources are cleaned up
    """

    def __init__(self, provider, **kwargs):
        super().__init__(**kwargs)
        self.provider = provider
        self.opts = provider.opts
        self.threads = {}

    def run(self):
        """
        Start the test process to kill
        """
        self.provider.lock()
        lockfile = self.provider._get_lock_file()
        log.debug("KillProcessTest acquried lock file %s", lockfile)

        killtest_pid = os.getpid()

        # check that lock has been released
        assert self.provider._master_lock.acquire(timeout=5)

        while True:
            tsleep = 1
            time.sleep(tsleep)  # give time for kill by sigterm


@pytest.mark.slow_test
@pytest.mark.skip_unless_on_linux
@pytest.mark.timeout_unless_on_windows(120)
def test_git_provider_sigterm_cleanup(main_class):
    """
    Start process which will obtain lock, and leave it locked
    then kill the process via SIGTERM and ensure locked resources are cleaned up
    """
    provider = main_class.remotes[0]

    with salt.utils.process.default_signals(signal.SIGINT, signal.SIGTERM):
        procmgr = salt.utils.process.ProcessManager(wait_for_kill=1)
        proc = procmgr.add_process(KillProcessTest, args=(provider,), name="test_kill")

    while not proc.is_alive():
        time.sleep(1)  # give some time for it to be started

    procmgr.run(asynchronous=True)

    time.sleep(2)  # give some time for it to terminate

    # child process should be alive
    file_name = provider._get_lock_file("update")

    assert pathlib.Path(file_name).exists()
    assert pathlib.Path(file_name).is_file()

    procmgr.terminate()  # sends a SIGTERM

    time.sleep(2)  # give some time for it to terminate

    assert not proc.is_alive()
    assert not pathlib.Path(file_name).exists()



"""
This runner is designed to mirror the execution module config.py, but for
master settings
"""

import salt.utils.data
import salt.utils.sdb


def get(key, default="", delimiter=":"):
    """
    Retrieve master config options, with optional nesting via the delimiter
    argument.

    **Arguments**

    default

        If the key is not found, the default will be returned instead

    delimiter

        Override the delimiter used to separate nested levels of a data
        structure.

    CLI Example:

    .. code-block:: bash

        salt-run config.get gitfs_remotes
        salt-run config.get file_roots:base
        salt-run config.get file_roots,base delimiter=','
    """
    ret = salt.utils.data.traverse_dict_and_list(
        __opts__, key, default="_|-", delimiter=delimiter
    )
    if ret == "_|-":
        return default
    else:
        return salt.utils.sdb.sdb_get(ret, __opts__)



"""
A state module to manage Palo Alto network devices.

:codeauthor: ``Spencer Ervin <spencer_ervin@hotmail.com>``
:maturity:   new
:depends:    none
:platform:   unix


About
=====
This state module was designed to handle connections to a Palo Alto based
firewall. This module relies on the Palo Alto proxy module to interface with the devices.

This state module is designed to give extreme flexibility in the control over XPATH values on the PANOS device. It
exposes the core XML API commands and allows state modules to chain complex XPATH commands.

Below is an example of how to construct a security rule and move to the top of the policy. This will take a config
lock to prevent execution during the operation, then remove the lock. After the XPATH has been deployed, it will
commit to the device.

.. code-block:: yaml

    panos/takelock:
        panos.add_config_lock
    panos/service_tcp_22:
        panos.set_config:
            - xpath: /config/devices/entry[@name='localhost.localdomain']/vsys/entry[@name='vsys1']/service
            - value: <entry name='tcp-22'><protocol><tcp><port>22</port></tcp></protocol></entry>
            - commit: False
    panos/create_rule1:
        panos.set_config:
            - xpath: /config/devices/entry[@name='localhost.localdomain']/vsys/entry[@name='vsys1']/rulebase/security/rules
            - value: '
              <entry name="rule1">
                <from><member>trust</member></from>
                <to><member>untrust</member></to>
                <source><member>10.0.0.1</member></source>
                <destination><member>10.0.1.1</member></destination>
                <service><member>tcp-22</member></service>
                <application><member>any</member></application>
                <action>allow</action>
                <disabled>no</disabled>
              </entry>'
            - commit: False
    panos/moveruletop:
        panos.move_config:
            - xpath: /config/devices/entry[@name='localhost.localdomain']/vsys/entry[@name='vsys1']/rulebase/security/rules/entry[@name='rule1']
            - where: top
            - commit: False
    panos/removelock:
        panos.remove_config_lock
    panos/commit:
        panos.commit_config

Version Specific Configurations
===============================
Palo Alto devices running different versions will have different supported features and different command structures. In
order to account for this, the proxy module can be leveraged to check if the panos device is at a specific revision
level.

The proxy['panos.is_required_version'] method will check if a panos device is currently running a version equal or
greater than the passed version. For example, proxy['panos.is_required_version']('7.0.0') would match both 7.1.0 and
8.0.0.

.. code-block:: jinja

    {% if proxy['panos.is_required_version']('8.0.0') %}
    panos/deviceconfig/system/motd-and-banner:
      panos.set_config:
        - xpath: /config/devices/entry[@name='localhost.localdomain']/deviceconfig/system/motd-and-banner
        - value: |
          <banner-header>BANNER TEXT</banner-header>
          <banner-header-color>color2</banner-header-color>
          <banner-header-text-color>color18</banner-header-text-color>
          <banner-header-footer-match>yes</banner-header-footer-match>
        - commit: False
    {% endif %}

.. seealso::
    :py:mod:`Palo Alto Proxy Module <salt.proxy.panos>`

"""

import logging
import xml.etree.ElementTree as ET

import salt.utils.xmlutil as xml

log = logging.getLogger(__name__)


def __virtual__():
    if "panos.commit" in __salt__:
        return True
    return (False, "panos module could not be loaded")


def _build_members(members, anycheck=False):
    """
    Builds a member formatted string for XML operation.

    """
    if isinstance(members, list):

        # This check will strip down members to a single any statement
        if anycheck and "any" in members:
            return "<member>any</member>"
        response = ""
        for m in members:
            response += f"<member>{m}</member>"
        return response
    else:
        return f"<member>{members}</member>"


def _default_ret(name):
    """
    Set the default response values.

    """
    ret = {"name": name, "changes": {}, "commit": None, "result": False, "comment": ""}
    return ret


def _edit_config(xpath, element):
    """
    Sends an edit request to the device.

    """
    query = {"type": "config", "action": "edit", "xpath": xpath, "element": element}

    response = __proxy__["panos.call"](query)

    return _validate_response(response)


def _get_config(xpath):
    """
    Retrieves an xpath from the device.

    """
    query = {"type": "config", "action": "get", "xpath": xpath}

    response = __proxy__["panos.call"](query)

    return response


def _move_after(xpath, target):
    """
    Moves an xpath to the after of its section.

    """
    query = {
        "type": "config",
        "action": "move",
        "xpath": xpath,
        "where": "after",
        "dst": target,
    }

    response = __proxy__["panos.call"](query)

    return _validate_response(response)


def _move_before(xpath, target):
    """
    Moves an xpath to the bottom of its section.

    """
    query = {
        "type": "config",
        "action": "move",
        "xpath": xpath,
        "where": "before",
        "dst": target,
    }

    response = __proxy__["panos.call"](query)

    return _validate_response(response)


def _move_bottom(xpath):
    """
    Moves an xpath to the bottom of its section.

    """
    query = {"type": "config", "action": "move", "xpath": xpath, "where": "bottom"}

    response = __proxy__["panos.call"](query)

    return _validate_response(response)


def _move_top(xpath):
    """
    Moves an xpath to the top of its section.

    """
    query = {"type": "config", "action": "move", "xpath": xpath, "where": "top"}

    response = __proxy__["panos.call"](query)

    return _validate_response(response)


def _set_config(xpath, element):
    """
    Sends a set request to the device.

    """
    query = {"type": "config", "action": "set", "xpath": xpath, "element": element}

    response = __proxy__["panos.call"](query)

    return _validate_response(response)


def _validate_response(response):
    """
    Validates a response from a Palo Alto device. Used to verify success of commands.

    """
    if not response:
        return False, "Unable to validate response from device."
    elif "msg" in response:
        if "line" in response["msg"]:
            if response["msg"]["line"] == "already at the top":
                return True, response
            elif response["msg"]["line"] == "already at the bottom":
                return True, response
            else:
                return False, response
        elif response["msg"] == "command succeeded":
            return True, response
        else:
            return False, response
    elif "status" in response:
        if response["status"] == "success":
            return True, response
        else:
            return False, response
    else:
        return False, response


def add_config_lock(name):
    """
    Prevent other users from changing configuration until the lock is released.

    name: The name of the module function to execute.

    SLS Example:

    .. code-block:: yaml

        panos/takelock:
            panos.add_config_lock

    """
    ret = _default_ret(name)

    ret.update({"changes": __salt__["panos.add_config_lock"](), "result": True})

    return ret


def address_exists(
    name,
    addressname=None,
    vsys=1,
    ipnetmask=None,
    iprange=None,
    fqdn=None,
    description=None,
    commit=False,
):
    """
    Ensures that an address object exists in the configured state. If it does not exist or is not configured with the
    specified attributes, it will be adjusted to match the specified values.

    This module will only process a single address type (ip-netmask, ip-range, or fqdn). It will process the specified
    value if the following order: ip-netmask, ip-range, fqdn. For proper execution, only specify a single address
    type.

    name: The name of the module function to execute.

    addressname(str): The name of the address object.  The name is case-sensitive and can have up to 31 characters,
    which an be letters, numbers, spaces, hyphens, and underscores. The name must be unique on a firewall and, on
    Panorama, unique within its device group and any ancestor or descendant device groups.

    vsys(str): The string representation of the VSYS ID. Defaults to VSYS 1.

    ipnetmask(str): The IPv4 or IPv6 address or IP address range using the format ip_address/mask or ip_address where
    the mask is the number of significant binary digits used for the network portion of the address. Ideally, for IPv6,
    you specify only the network portion, not the host portion.

    iprange(str): A range of addresses using the format ip_address–ip_address where both addresses can be  IPv4 or both
    can be IPv6.

    fqdn(str): A fully qualified domain name format. The FQDN initially resolves at commit time. Entries are
    subsequently refreshed when the firewall performs a check every 30 minutes; all changes in the IP address for the
    entries are picked up at the refresh cycle.

    description(str): A description for the policy (up to 255 characters).

    commit(bool): If true the firewall will commit the changes, if false do not commit changes.

    SLS Example:

    .. code-block:: yaml

        panos/address/h-10.10.10.10:
            panos.address_exists:
              - addressname: h-10.10.10.10
              - vsys: 1
              - ipnetmask: 10.10.10.10
              - commit: False

        panos/address/10.0.0.1-10.0.0.50:
            panos.address_exists:
              - addressname: r-10.0.0.1-10.0.0.50
              - vsys: 1
              - iprange: 10.0.0.1-10.0.0.50
              - commit: False

        panos/address/foo.bar.com:
            panos.address_exists:
              - addressname: foo.bar.com
              - vsys: 1
              - fqdn: foo.bar.com
              - description: My fqdn object
              - commit: False

    """
    ret = _default_ret(name)

    if not addressname:
        ret.update({"comment": "The service name field must be provided."})
        return ret

    # Check if address object currently exists
    address = __salt__["panos.get_address"](addressname, vsys)["result"]

    if address and "entry" in address:
        address = address["entry"]
    else:
        address = {}

    element = ""

    # Verify the arguments
    if ipnetmask:
        element = f"<ip-netmask>{ipnetmask}</ip-netmask>"
    elif iprange:
        element = f"<ip-range>{iprange}</ip-range>"
    elif fqdn:
        element = f"<fqdn>{fqdn}</fqdn>"
    else:
        ret.update({"comment": "A valid address type must be specified."})
        return ret

    if description:
        element += f"<description>{description}</description>"

    full_element = f"<entry name='{addressname}'>{element}</entry>"

    new_address = xml.to_dict(ET.fromstring(full_element), True)

    if address == new_address:
        ret.update(
            {
                "comment": "Address object already exists. No changes required.",
                "result": True,
            }
        )
        return ret
    else:
        xpath = (
            "/config/devices/entry[@name='localhost.localdomain']/vsys/entry[@name='vsys{}']/address/"
            "entry[@name='{}']".format(vsys, addressname)
        )

        result, msg = _edit_config(xpath, full_element)

        if not result:
            ret.update({"comment": msg})
            return ret

    if commit is True:
        ret.update(
            {
                "changes": {"before": address, "after": new_address},
                "commit": __salt__["panos.commit"](),
                "comment": "Address object successfully configured.",
                "result": True,
            }
        )
    else:
        ret.update(
            {
                "changes": {"before": address, "after": new_address},
                "comment": "Service object successfully configured.",
                "result": True,
            }
        )

    return ret


def address_group_exists(
    name, groupname=None, vsys=1, members=None, description=None, commit=False
):
    """
    Ensures that an address group object exists in the configured state. If it does not exist or is not configured with
    the specified attributes, it will be adjusted to match the specified values.

    This module will enforce group membership. If a group exists and contains members this state does not include,
    those members will be removed and replaced with the specified members in the state.

    name: The name of the module function to execute.

    groupname(str): The name of the address group object.  The name is case-sensitive and can have up to 31 characters,
    which an be letters, numbers, spaces, hyphens, and underscores. The name must be unique on a firewall and, on
    Panorama, unique within its device group and any ancestor or descendant device groups.

    vsys(str): The string representation of the VSYS ID. Defaults to VSYS 1.

    members(str, list): The members of the address group. These must be valid address objects or address groups on the
    system that already exist prior to the execution of this state.

    description(str): A description for the policy (up to 255 characters).

    commit(bool): If true the firewall will commit the changes, if false do not commit changes.

    SLS Example:

    .. code-block:: yaml

        panos/address-group/my-group:
            panos.address_group_exists:
              - groupname: my-group
              - vsys: 1
              - members:
                - my-address-object
                - my-other-address-group
              - description: A group that needs to exist
              - commit: False

    """
    ret = _default_ret(name)

    if not groupname:
        ret.update({"comment": "The group name field must be provided."})
        return ret

    # Check if address group object currently exists
    group = __salt__["panos.get_address_group"](groupname, vsys)["result"]

    if group and "entry" in group:
        group = group["entry"]
    else:
        group = {}

    # Verify the arguments
    if members:
        element = f"<static>{_build_members(members, True)}</static>"
    else:
        ret.update({"comment": "The group members must be provided."})
        return ret

    if description:
        element += f"<description>{description}</description>"

    full_element = f"<entry name='{groupname}'>{element}</entry>"

    new_group = xml.to_dict(ET.fromstring(full_element), True)

    if group == new_group:
        ret.update(
            {
                "comment": "Address group object already exists. No changes required.",
                "result": True,
            }
        )
        return ret
    else:
        xpath = (
            "/config/devices/entry[@name='localhost.localdomain']/vsys/entry[@name='vsys{}']/address-group/"
            "entry[@name='{}']".format(vsys, groupname)
        )

        result, msg = _edit_config(xpath, full_element)

        if not result:
            ret.update({"comment": msg})
            return ret

    if commit is True:
        ret.update(
            {
                "changes": {"before": group, "after": new_group},
                "commit": __salt__["panos.commit"](),
                "comment": "Address group object successfully configured.",
                "result": True,
            }
        )
    else:
        ret.update(
            {
                "changes": {"before": group, "after": new_group},
                "comment": "Address group object successfully configured.",
                "result": True,
            }
        )

    return ret


def clone_config(name, xpath=None, newname=None, commit=False):
    """
    Clone a specific XPATH and set it to a new name.

    name: The name of the module function to execute.

    xpath(str): The XPATH of the configuration API tree to clone.

    newname(str): The new name of the XPATH clone.

    commit(bool): If true the firewall will commit the changes, if false do not commit changes.

    SLS Example:

    .. code-block:: yaml

        panos/clonerule:
            panos.clone_config:
              - xpath: /config/devices/entry/vsys/entry[@name='vsys1']/rulebase/security/rules&from=/config/devices/
              entry/vsys/entry[@name='vsys1']/rulebase/security/rules/entry[@name='rule1']
              - value: rule2
              - commit: True

    """
    ret = _default_ret(name)

    if not xpath:
        return ret

    if not newname:
        return ret

    query = {"type": "config", "action": "clone", "xpath": xpath, "newname": newname}

    result, response = _validate_response(__proxy__["panos.call"](query))

    ret.update({"changes": response, "result": result})

    if not result:
        return ret

    if commit is True:
        ret.update({"commit": __salt__["panos.commit"](), "result": True})

    return ret


def commit_config(name):
    """
    Commits the candidate configuration to the running configuration.

    name: The name of the module function to execute.

    SLS Example:

    .. code-block:: yaml

        panos/commit:
            panos.commit_config

    """
    ret = _default_ret(name)

    ret.update({"commit": __salt__["panos.commit"](), "result": True})

    return ret


def delete_config(name, xpath=None, commit=False):
    """
    Deletes a Palo Alto XPATH to a specific value.

    Use the xpath parameter to specify the location of the object to be deleted.

    name: The name of the module function to execute.

    xpath(str): The XPATH of the configuration API tree to control.

    commit(bool): If true the firewall will commit the changes, if false do not commit changes.

    SLS Example:

    .. code-block:: yaml

        panos/deletegroup:
            panos.delete_config:
              - xpath: /config/devices/entry/vsys/entry[@name='vsys1']/address-group/entry[@name='test']
              - commit: True

    """
    ret = _default_ret(name)

    if not xpath:
        return ret

    query = {"type": "config", "action": "delete", "xpath": xpath}

    result, response = _validate_response(__proxy__["panos.call"](query))

    ret.update({"changes": response, "result": result})

    if not result:
        return ret

    if commit is True:
        ret.update({"commit": __salt__["panos.commit"](), "result": True})

    return ret


def download_software(name, version=None, synch=False, check=False):
    """
    Ensures that a software version is downloaded.

    name: The name of the module function to execute.

    version(str): The software version to check. If this version is not already downloaded, it will attempt to download
    the file from Palo Alto.

    synch(bool): If true, after downloading the file it will be synched to its peer.

    check(bool): If true, the PANOS device will first attempt to pull the most recent software inventory list from Palo
    Alto.

    SLS Example:

    .. code-block:: yaml

        panos/version8.0.0:
            panos.download_software:
              - version: 8.0.0
              - synch: False
              - check: True

    """
    ret = _default_ret(name)

    if check is True:
        __salt__["panos.check_software"]()

    versions = __salt__["panos.get_software_info"]()["result"]

    if (
        "sw-updates" not in versions
        or "versions" not in versions["sw-updates"]
        or "entry" not in versions["sw-updates"]["versions"]
    ):
        ret.update(
            {
                "comment": "Software version is not found in the local software list.",
                "result": False,
            }
        )
        return ret

    for entry in versions["sw-updates"]["versions"]["entry"]:
        if entry["version"] == version and entry["downloaded"] == "yes":
            ret.update(
                {"comment": "Software version is already downloaded.", "result": True}
            )
        return ret

    ret.update(
        {
            "changes": __salt__["panos.download_software_version"](
                version=version, synch=synch
            )
        }
    )

    versions = __salt__["panos.get_software_info"]()["result"]

    if (
        "sw-updates" not in versions
        or "versions" not in versions["sw-updates"]
        or "entry" not in versions["sw-updates"]["versions"]
    ):
        ret.update({"result": False})
        return ret

    for entry in versions["sw-updates"]["versions"]["entry"]:
        if entry["version"] == version and entry["downloaded"] == "yes":
            ret.update({"result": True})
        return ret

    return ret


def edit_config(name, xpath=None, value=None, commit=False):
    """
    Edits a Palo Alto XPATH to a specific value. This will always overwrite the existing value, even if it is not
    changed.

    You can replace an existing object hierarchy at a specified location in the configuration with a new value. Use
    the xpath parameter to specify the location of the object, including the node to be replaced.

    This is the recommended state to enforce configurations on a xpath.

    name: The name of the module function to execute.

    xpath(str): The XPATH of the configuration API tree to control.

    value(str): The XML value to edit. This must be a child to the XPATH.

    commit(bool): If true the firewall will commit the changes, if false do not commit changes.

    SLS Example:

    .. code-block:: yaml

        panos/addressgroup:
            panos.edit_config:
              - xpath: /config/devices/entry/vsys/entry[@name='vsys1']/address-group/entry[@name='test']
              - value: <static><entry name='test'><member>abc</member><member>xyz</member></entry></static>
              - commit: True

    """
    ret = _default_ret(name)

    # Verify if the current XPATH is equal to the specified value.
    # If we are equal, no changes required.
    xpath_split = xpath.split("/")

    # Retrieve the head of the xpath for validation.
    if len(xpath_split) > 0:
        head = xpath_split[-1]
        if "[" in head:
            head = head.split("[")[0]

    current_element = __salt__["panos.get_xpath"](xpath)["result"]

    if head and current_element and head in current_element:
        current_element = current_element[head]
    else:
        current_element = {}

    new_element = xml.to_dict(ET.fromstring(value), True)

    if current_element == new_element:
        ret.update(
            {
                "comment": "XPATH is already equal to the specified value.",
                "result": True,
            }
        )
        return ret

    result, msg = _edit_config(xpath, value)

    ret.update({"comment": msg, "result": result})

    if not result:
        return ret

    if commit is True:
        ret.update(
            {
                "changes": {"before": current_element, "after": new_element},
                "commit": __salt__["panos.commit"](),
                "result": True,
            }
        )
    else:
        ret.update(
            {
                "changes": {"before": current_element, "after": new_element},
                "result": True,
            }
        )

    return ret


def move_config(name, xpath=None, where=None, dst=None, commit=False):
    """
    Moves a XPATH value to a new location.

    Use the xpath parameter to specify the location of the object to be moved, the where parameter to
    specify type of move, and dst parameter to specify the destination path.

    name: The name of the module function to execute.

    xpath(str): The XPATH of the configuration API tree to move.

    where(str): The type of move to execute. Valid options are after, before, top, bottom. The after and before
    options will require the dst option to specify the destination of the action. The top action will move the
    XPATH to the top of its structure. The botoom action will move the XPATH to the bottom of its structure.

    dst(str): Optional. Specifies the destination to utilize for a move action. This is ignored for the top
    or bottom action.

    commit(bool): If true the firewall will commit the changes, if false do not commit changes. If the operation is
    not successful, it will not commit.

    SLS Example:

    .. code-block:: yaml

        panos/moveruletop:
            panos.move_config:
              - xpath: /config/devices/entry/vsys/entry[@name='vsys1']/rulebase/security/rules/entry[@name='rule1']
              - where: top
              - commit: True

        panos/moveruleafter:
            panos.move_config:
              - xpath: /config/devices/entry/vsys/entry[@name='vsys1']/rulebase/security/rules/entry[@name='rule1']
              - where: after
              - dst: rule2
              - commit: True

    """
    ret = _default_ret(name)

    if not xpath:
        return ret

    if not where:
        return ret

    if where == "after":
        result, msg = _move_after(xpath, dst)
    elif where == "before":
        result, msg = _move_before(xpath, dst)
    elif where == "top":
        result, msg = _move_top(xpath)
    elif where == "bottom":
        result, msg = _move_bottom(xpath)

    ret.update({"result": result, "comment": msg})

    if not result:
        return ret

    if commit is True:
        ret.update({"commit": __salt__["panos.commit"](), "result": True})

    return ret


def remove_config_lock(name):
    """
    Release config lock previously held.

    name: The name of the module function to execute.

    SLS Example:

    .. code-block:: yaml

        panos/takelock:
            panos.remove_config_lock

    """
    ret = _default_ret(name)

    ret.update({"changes": __salt__["panos.remove_config_lock"](), "result": True})

    return ret


def rename_config(name, xpath=None, newname=None, commit=False):
    """
    Rename a Palo Alto XPATH to a specific value. This will always rename the value even if a change is not needed.

    name: The name of the module function to execute.

    xpath(str): The XPATH of the configuration API tree to control.

    newname(str): The new name of the XPATH value.

    commit(bool): If true the firewall will commit the changes, if false do not commit changes.

    SLS Example:

    .. code-block:: yaml

        panos/renamegroup:
            panos.rename_config:
              - xpath: /config/devices/entry/vsys/entry[@name='vsys1']/address/entry[@name='old_address']
              - value: new_address
              - commit: True

    """
    ret = _default_ret(name)

    if not xpath:
        return ret

    if not newname:
        return ret

    query = {"type": "config", "action": "rename", "xpath": xpath, "newname": newname}

    result, response = _validate_response(__proxy__["panos.call"](query))

    ret.update({"changes": response, "result": result})

    if not result:
        return ret

    if commit is True:
        ret.update({"commit": __salt__["panos.commit"](), "result": True})

    return ret


def security_rule_exists(
    name,
    rulename=None,
    vsys="1",
    action=None,
    disabled=None,
    sourcezone=None,
    destinationzone=None,
    source=None,
    destination=None,
    application=None,
    service=None,
    description=None,
    logsetting=None,
    logstart=None,
    logend=None,
    negatesource=None,
    negatedestination=None,
    profilegroup=None,
    datafilter=None,
    fileblock=None,
    spyware=None,
    urlfilter=None,
    virus=None,
    vulnerability=None,
    wildfire=None,
    move=None,
    movetarget=None,
    commit=False,
):
    """
    Ensures that a security rule exists on the device. Also, ensure that all configurations are set appropriately.

    This method will create the rule if it does not exist. If the rule does exist, it will ensure that the
    configurations are set appropriately.

    If the rule does not exist and is created, any value that is not provided will be provided as the default.
    The action, to, from, source, destination, application, and service fields are mandatory and must be provided.

    This will enforce the exact match of the rule. For example, if the rule is currently configured with the log-end
    option, but this option is not specified in the state method, it will be removed and reset to the system default.

    It is strongly recommended to specify all options to ensure proper operation.

    When defining the profile group settings, the device can only support either a profile group or individual settings.
    If both are specified, the profile group will be preferred and the individual settings are ignored. If neither are
    specified, the value will be set to system default of none.

    name: The name of the module function to execute.

    rulename(str): The name of the security rule.  The name is case-sensitive and can have up to 31 characters, which
    can be letters, numbers, spaces, hyphens, and underscores. The name must be unique on a firewall and, on Panorama,
    unique within its device group and any ancestor or descendant device groups.

    vsys(str): The string representation of the VSYS ID. Defaults to VSYS 1.

    action(str): The action that the security rule will enforce. Valid options are: allow, deny, drop, reset-client,
    reset-server, reset-both.

    disabled(bool): Controls if the rule is disabled. Set 'True' to disable and 'False' to enable.

    sourcezone(str, list): The source zone(s). The value 'any' will match all zones.

    destinationzone(str, list): The destination zone(s). The value 'any' will match all zones.

    source(str, list): The source address(es). The value 'any' will match all addresses.

    destination(str, list): The destination address(es). The value 'any' will match all addresses.

    application(str, list): The application(s) matched. The value 'any' will match all applications.

    service(str, list): The service(s) matched. The value 'any' will match all services. The value
    'application-default' will match based upon the application defined ports.

    description(str): A description for the policy (up to 255 characters).

    logsetting(str): The name of a valid log forwarding profile.

    logstart(bool): Generates a traffic log entry for the start of a session (disabled by default).

    logend(bool): Generates a traffic log entry for the end of a session (enabled by default).

    negatesource(bool): Match all but the specified source addresses.

    negatedestination(bool): Match all but the specified destination addresses.

    profilegroup(str): A valid profile group name.

    datafilter(str): A valid data filter profile name. Ignored with the profilegroup option set.

    fileblock(str): A valid file blocking profile name. Ignored with the profilegroup option set.

    spyware(str): A valid spyware profile name. Ignored with the profilegroup option set.

    urlfilter(str): A valid URL filtering profile name. Ignored with the profilegroup option set.

    virus(str): A valid virus profile name. Ignored with the profilegroup option set.

    vulnerability(str): A valid vulnerability profile name. Ignored with the profilegroup option set.

    wildfire(str): A valid vulnerability profile name. Ignored with the profilegroup option set.

    move(str): An optional argument that ensure the rule is moved to a specific location. Valid options are 'top',
    'bottom', 'before', or 'after'. The 'before' and 'after' options require the use of the 'movetarget' argument
    to define the location of the move request.

    movetarget(str): An optional argument that defines the target of the move operation if the move argument is
    set to 'before' or 'after'.

    commit(bool): If true the firewall will commit the changes, if false do not commit changes.

    SLS Example:

    .. code-block:: yaml

        panos/rulebase/security/rule01:
            panos.security_rule_exists:
              - rulename: rule01
              - vsys: 1
              - action: allow
              - disabled: False
              - sourcezone: untrust
              - destinationzone: trust
              - source:
                - 10.10.10.0/24
                - 1.1.1.1
              - destination:
                - 2.2.2.2-2.2.2.4
              - application:
                - any
              - service:
                - tcp-25
              - description: My test security rule
              - logsetting: logprofile
              - logstart: False
              - logend: True
              - negatesource: False
              - negatedestination: False
              - profilegroup: myprofilegroup
              - move: top
              - commit: False

        panos/rulebase/security/rule01:
            panos.security_rule_exists:
              - rulename: rule01
              - vsys: 1
              - action: allow
              - disabled: False
              - sourcezone: untrust
              - destinationzone: trust
              - source:
                - 10.10.10.0/24
                - 1.1.1.1
              - destination:
                - 2.2.2.2-2.2.2.4
              - application:
                - any
              - service:
                - tcp-25
              - description: My test security rule
              - logsetting: logprofile
              - logstart: False
              - logend: False
              - datafilter: foobar
              - fileblock: foobar
              - spyware: foobar
              - urlfilter: foobar
              - virus: foobar
              - vulnerability: foobar
              - wildfire: foobar
              - move: after
              - movetarget: rule02
              - commit: False
    """
    ret = _default_ret(name)

    if not rulename:
        return ret

    # Check if rule currently exists
    rule = __salt__["panos.get_security_rule"](rulename, vsys)["result"]

    if rule and "entry" in rule:
        rule = rule["entry"]
    else:
        rule = {}

    # Build the rule element
    element = ""
    if sourcezone:
        element += f"<from>{_build_members(sourcezone, True)}</from>"
    else:
        ret.update({"comment": "The sourcezone field must be provided."})
        return ret

    if destinationzone:
        element += f"<to>{_build_members(destinationzone, True)}</to>"
    else:
        ret.update({"comment": "The destinationzone field must be provided."})
        return ret

    if source:
        element += f"<source>{_build_members(source, True)}</source>"
    else:
        ret.update({"comment": "The source field must be provided."})
        return

    if destination:
        element += "<destination>{}</destination>".format(
            _build_members(destination, True)
        )
    else:
        ret.update({"comment": "The destination field must be provided."})
        return ret

    if application:
        element += "<application>{}</application>".format(
            _build_members(application, True)
        )
    else:
        ret.update({"comment": "The application field must be provided."})
        return ret

    if service:
        element += f"<service>{_build_members(service, True)}</service>"
    else:
        ret.update({"comment": "The service field must be provided."})
        return ret

    if action:
        element += f"<action>{action}</action>"
    else:
        ret.update({"comment": "The action field must be provided."})
        return ret

    if disabled is not None:
        if disabled:
            element += "<disabled>yes</disabled>"
        else:
            element += "<disabled>no</disabled>"

    if description:
        element += f"<description>{description}</description>"

    if logsetting:
        element += f"<log-setting>{logsetting}</log-setting>"

    if logstart is not None:
        if logstart:
            element += "<log-start>yes</log-start>"
        else:
            element += "<log-start>no</log-start>"

    if logend is not None:
        if logend:
            element += "<log-end>yes</log-end>"
        else:
            element += "<log-end>no</log-end>"

    if negatesource is not None:
        if negatesource:
            element += "<negate-source>yes</negate-source>"
        else:
            element += "<negate-source>no</negate-source>"

    if negatedestination is not None:
        if negatedestination:
            element += "<negate-destination>yes</negate-destination>"
        else:
            element += "<negate-destination>no</negate-destination>"

    # Build the profile settings
    profile_string = None
    if profilegroup:
        profile_string = f"<group><member>{profilegroup}</member></group>"
    else:
        member_string = ""
        if datafilter:
            member_string += (
                "<data-filtering><member>{}</member></data-filtering>".format(
                    datafilter
                )
            )
        if fileblock:
            member_string += (
                f"<file-blocking><member>{fileblock}</member></file-blocking>"
            )
        if spyware:
            member_string += f"<spyware><member>{spyware}</member></spyware>"
        if urlfilter:
            member_string += (
                f"<url-filtering><member>{urlfilter}</member></url-filtering>"
            )
        if virus:
            member_string += f"<virus><member>{virus}</member></virus>"
        if vulnerability:
            member_string += (
                "<vulnerability><member>{}</member></vulnerability>".format(
                    vulnerability
                )
            )
        if wildfire:
            member_string += (
                "<wildfire-analysis><member>{}</member></wildfire-analysis>".format(
                    wildfire
                )
            )
        if member_string != "":
            profile_string = f"<profiles>{member_string}</profiles>"

    if profile_string:
        element += f"<profile-setting>{profile_string}</profile-setting>"

    full_element = f"<entry name='{rulename}'>{element}</entry>"

    new_rule = xml.to_dict(ET.fromstring(full_element), True)

    config_change = False

    if rule == new_rule:
        ret.update({"comment": "Security rule already exists. No changes required."})
    else:
        config_change = True
        xpath = (
            "/config/devices/entry[@name='localhost.localdomain']/vsys/entry[@name='vsys{}']/rulebase/"
            "security/rules/entry[@name='{}']".format(vsys, rulename)
        )

        result, msg = _edit_config(xpath, full_element)

        if not result:
            ret.update({"comment": msg})
            return ret

        ret.update(
            {
                "changes": {"before": rule, "after": new_rule},
                "comment": "Security rule verified successfully.",
            }
        )

    if move:
        movepath = (
            "/config/devices/entry[@name='localhost.localdomain']/vsys/entry[@name='vsys{}']/rulebase/"
            "security/rules/entry[@name='{}']".format(vsys, rulename)
        )
        move_result = False
        move_msg = ""
        if move == "before" and movetarget:
            move_result, move_msg = _move_before(movepath, movetarget)
        elif move == "after":
            move_result, move_msg = _move_after(movepath, movetarget)
        elif move == "top":
            move_result, move_msg = _move_top(movepath)
        elif move == "bottom":
            move_result, move_msg = _move_bottom(movepath)

        if config_change:
            ret.update(
                {"changes": {"before": rule, "after": new_rule, "move": move_msg}}
            )
        else:
            ret.update({"changes": {"move": move_msg}})

        if not move_result:
            ret.update({"comment": move_msg})
            return ret

    if commit is True:
        ret.update({"commit": __salt__["panos.commit"](), "result": True})
    else:
        ret.update({"result": True})

    return ret


def service_exists(
    name,
    servicename=None,
    vsys=1,
    protocol=None,
    port=None,
    description=None,
    commit=False,
):
    """
    Ensures that a service object exists in the configured state. If it does not exist or is not configured with the
    specified attributes, it will be adjusted to match the specified values.

    name: The name of the module function to execute.

    servicename(str): The name of the security object.  The name is case-sensitive and can have up to 31 characters,
    which an be letters, numbers, spaces, hyphens, and underscores. The name must be unique on a firewall and, on
    Panorama, unique within its device group and any ancestor or descendant device groups.

    vsys(str): The string representation of the VSYS ID. Defaults to VSYS 1.

    protocol(str): The protocol that is used by the service object. The only valid options are tcp and udp.

    port(str): The port number that is used by the service object. This can be specified as a single integer or a
    valid range of ports.

    description(str): A description for the policy (up to 255 characters).

    commit(bool): If true the firewall will commit the changes, if false do not commit changes.

    SLS Example:

    .. code-block:: yaml

        panos/service/tcp-80:
            panos.service_exists:
              - servicename: tcp-80
              - vsys: 1
              - protocol: tcp
              - port: 80
              - description: Hypertext Transfer Protocol
              - commit: False

        panos/service/udp-500-550:
            panos.service_exists:
              - servicename: udp-500-550
              - vsys: 3
              - protocol: udp
              - port: 500-550
              - commit: False

    """
    ret = _default_ret(name)

    if not servicename:
        ret.update({"comment": "The service name field must be provided."})
        return ret

    # Check if service object currently exists
    service = __salt__["panos.get_service"](servicename, vsys)["result"]

    if service and "entry" in service:
        service = service["entry"]
    else:
        service = {}

    # Verify the arguments
    if not protocol and protocol not in ["tcp", "udp"]:
        ret.update({"comment": "The protocol must be provided and must be tcp or udp."})
        return ret
    if not port:
        ret.update({"comment": "The port field must be provided."})
        return ret

    element = "<protocol><{0}><port>{1}</port></{0}></protocol>".format(protocol, port)

    if description:
        element += f"<description>{description}</description>"

    full_element = f"<entry name='{servicename}'>{element}</entry>"

    new_service = xml.to_dict(ET.fromstring(full_element), True)

    if service == new_service:
        ret.update(
            {
                "comment": "Service object already exists. No changes required.",
                "result": True,
            }
        )
        return ret
    else:
        xpath = (
            "/config/devices/entry[@name='localhost.localdomain']/vsys/entry[@name='vsys{}']/service/"
            "entry[@name='{}']".format(vsys, servicename)
        )

        result, msg = _edit_config(xpath, full_element)

        if not result:
            ret.update({"comment": msg})
            return ret

    if commit is True:
        ret.update(
            {
                "changes": {"before": service, "after": new_service},
                "commit": __salt__["panos.commit"](),
                "comment": "Service object successfully configured.",
                "result": True,
            }
        )
    else:
        ret.update(
            {
                "changes": {"before": service, "after": new_service},
                "comment": "Service object successfully configured.",
                "result": True,
            }
        )

    return ret


def service_group_exists(
    name, groupname=None, vsys=1, members=None, description=None, commit=False
):
    """
    Ensures that a service group object exists in the configured state. If it does not exist or is not configured with
    the specified attributes, it will be adjusted to match the specified values.

    This module will enforce group membership. If a group exists and contains members this state does not include,
    those members will be removed and replaced with the specified members in the state.

    name: The name of the module function to execute.

    groupname(str): The name of the service group object.  The name is case-sensitive and can have up to 31 characters,
    which an be letters, numbers, spaces, hyphens, and underscores. The name must be unique on a firewall and, on
    Panorama, unique within its device group and any ancestor or descendant device groups.

    vsys(str): The string representation of the VSYS ID. Defaults to VSYS 1.

    members(str, list): The members of the service group. These must be valid service objects or service groups on the
    system that already exist prior to the execution of this state.

    description(str): A description for the policy (up to 255 characters).

    commit(bool): If true the firewall will commit the changes, if false do not commit changes.

    SLS Example:

    .. code-block:: yaml

        panos/service-group/my-group:
            panos.service_group_exists:
              - groupname: my-group
              - vsys: 1
              - members:
                - tcp-80
                - custom-port-group
              - description: A group that needs to exist
              - commit: False

    """
    ret = _default_ret(name)

    if not groupname:
        ret.update({"comment": "The group name field must be provided."})
        return ret

    # Check if service group object currently exists
    group = __salt__["panos.get_service_group"](groupname, vsys)["result"]

    if group and "entry" in group:
        group = group["entry"]
    else:
        group = {}

    # Verify the arguments
    if members:
        element = f"<members>{_build_members(members, True)}</members>"
    else:
        ret.update({"comment": "The group members must be provided."})
        return ret

    if description:
        element += f"<description>{description}</description>"

    full_element = f"<entry name='{groupname}'>{element}</entry>"

    new_group = xml.to_dict(ET.fromstring(full_element), True)

    if group == new_group:
        ret.update(
            {
                "comment": "Service group object already exists. No changes required.",
                "result": True,
            }
        )
        return ret
    else:
        xpath = (
            "/config/devices/entry[@name='localhost.localdomain']/vsys/entry[@name='vsys{}']/service-group/"
            "entry[@name='{}']".format(vsys, groupname)
        )

        result, msg = _edit_config(xpath, full_element)

        if not result:
            ret.update({"comment": msg})
            return ret

    if commit is True:
        ret.update(
            {
                "changes": {"before": group, "after": new_group},
                "commit": __salt__["panos.commit"](),
                "comment": "Service group object successfully configured.",
                "result": True,
            }
        )
    else:
        ret.update(
            {
                "changes": {"before": group, "after": new_group},
                "comment": "Service group object successfully configured.",
                "result": True,
            }
        )

    return ret


def set_config(name, xpath=None, value=None, commit=False):
    """
    Sets a Palo Alto XPATH to a specific value. This will always overwrite the existing value, even if it is not
    changed.

    You can add or create a new object at a specified location in the configuration hierarchy. Use the xpath parameter
    to specify the location of the object in the configuration

    name: The name of the module function to execute.

    xpath(str): The XPATH of the configuration API tree to control.

    value(str): The XML value to set. This must be a child to the XPATH.

    commit(bool): If true the firewall will commit the changes, if false do not commit changes.

    SLS Example:

    .. code-block:: yaml

        panos/hostname:
            panos.set_config:
              - xpath: /config/devices/entry[@name='localhost.localdomain']/deviceconfig/system
              - value: <hostname>foobar</hostname>
              - commit: True

    """
    ret = _default_ret(name)

    result, msg = _set_config(xpath, value)

    ret.update({"comment": msg, "result": result})

    if not result:
        return ret

    if commit is True:
        ret.update({"commit": __salt__["panos.commit"](), "result": True})

    return ret



"""
Module for OpenStack Management

:codeauthor:    Konrad Mosoń <mosonkonrad@gmail.com>
:maturity:      new
:depends:       openstack-utils
:platform:      linux
"""

import logging
import os.path

import salt.utils.files
import salt.utils.stringutils

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "openstack_mng"


def __virtual__():
    """
    Only load this module if openstack-service is installed
    """
    if os.path.isfile("/usr/bin/openstack-service"):
        return __virtualname__
    else:
        return (False, "The openstack-service binary could not be found.")


def start_service(service_name):
    """
    Start OpenStack service immediately

    CLI Example:

    .. code-block:: bash

        salt '*' openstack_mng.start_service neutron
    """

    os_cmd = ["/usr/bin/openstack-service", "start", service_name]
    return __salt__["cmd.retcode"](os_cmd) == 0


def stop_service(service_name):
    """
    Stop OpenStack service immediately

    CLI Example:

    .. code-block:: bash

        salt '*' openstack_mng.stop_service neutron
    """

    os_cmd = ["/usr/bin/openstack-service", "stop", service_name]
    return __salt__["cmd.retcode"](os_cmd) == 0


def restart_service(service_name, minimum_running_time=None):
    """
    Restart OpenStack service immediately, or only if it's running longer than
    specified value

    CLI Example:

    .. code-block:: bash

        salt '*' openstack_mng.restart_service neutron
        salt '*' openstack_mng.restart_service neutron minimum_running_time=600
    """

    if minimum_running_time:
        ret_code = False
        # get system services list for interesting openstack service
        services = __salt__["cmd.run"](
            ["/usr/bin/openstack-service", "list", service_name]
        ).split("\n")
        for service in services:
            service_info = __salt__["service.show"](service)
            with salt.utils.files.fopen("/proc/uptime") as rfh:
                boot_time = float(
                    salt.utils.stringutils.to_unicode(rfh.read()).split(" ")[0]
                )

            expr_time = (
                int(service_info.get("ExecMainStartTimestampMonotonic", 0)) / 1000000
                < boot_time - minimum_running_time
            )
            expr_active = service_info.get("ActiveState") == "active"

            if expr_time or not expr_active:
                # restart specific system service
                ret = __salt__["service.restart"](service)
                if ret:
                    ret_code = True

        return ret_code
    else:
        # just restart
        os_cmd = ["/usr/bin/openstack-service", "restart", service_name]
        return __salt__["cmd.retcode"](os_cmd) == 0



'''
{{module_name}} execution module
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

{{short_description}}

.. versionadded:: {{version}}

:configuration:

    .. code-block:: yaml
        <your example config>

'''
import logging

import salt.utils.compat

try:
    #  Import libs...
    {% if depending_libraries %}
    import {{depending_libraries}}
    {% endif %}
    HAS_LIBS = True
    MISSING_PACKAGE_REASON = None
except ImportError as ie:
    HAS_LIBS = False
    MISSING_PACKAGE_REASON = ie.message

log = logging.getLogger(__name__)

__virtualname__ = '{{virtual_name}}'


def __virtual__():
    '''
    Only load this module if dependencies is installed on this minion.
    '''
    if HAS_LIBS:
        return __virtualname__
    return (False,
            'The {{module_name}} execution module failed to load:'
            'import error - {0}.'.format(MISSING_PACKAGE_REASON))


def __init__(opts):
    #  Put logic here to instantiate underlying jobs/connections
    pass


def my_action(params):
    # Replace this with your actions
    pass



"""
Environment sdb Module

:maintainer:    SaltStack
:maturity:      New
:depends:       None
:platform:      all

This module allows access to environment variables using an ``sdb://`` URI.

Example configuration for this module:

.. code-block:: yaml

    osenv:
      driver: env

WARNING:
--------
OS environment variables will be available
to read via SDB.
Please make sure you don't have any sensitive data
in your environment variables!!

Example usage of sdb env module:

.. code-block:: jinja

    set some env var:
      cmd.run:
        - name: echo {{ salt['sdb.set']('sdb://osenv/foo', 'bar') }}
        - order: 1

    {% if salt['sdb.get']('sdb://osenv/foo') == 'bar' %}
    always-changes-and-succeeds:
      test.succeed_with_changes:
        - name: foo
    {% else %}
    always-changes-and-fails:
      test.fail_with_changes:
        - name: foo
    {% endif  %}

The above example will return success.

The ``env`` sdb module can also be used with salt cloud.
Assuming you have exported the environment variable named
``compute`` (and have ``osenv`` defined).
The example below will look for the salt cloud config key ``compute_name``
in the environment:

.. code-block:: yaml

  my-openstack-config:
    compute_name: sdb://osenv/compute
    ..snip

"""

import os

__func_alias__ = {"set_": "set"}


def __virtual__():
    """
    Always load
    """
    return True


def set_(key, value, profile=None):
    """
    Set a key/value pair
    """
    # pylint: disable=unused-argument
    return os.environ.setdefault(key, value)


def get(key, profile=None):
    """
    Get a value
    """
    # pylint: disable=unused-argument
    return os.environ.get(key)



import logging

import salt.utils.stringutils

try:
    import yamllint
    from yamllint import linter
    from yamllint.config import YamlLintConfig

    HAS_YAMLLINT = True
except ImportError:
    HAS_YAMLLINT = False

log = logging.getLogger(__name__)


def has_yamllint():
    """
    report if yamllint could be imported safly. allowing for clean import detection
    """
    return HAS_YAMLLINT


def version():
    """
    report version of yamllint installed for version comparison
    """
    return yamllint.__version__


def lint(
    source,
    yamlconf=None,
):
    """
    lint yaml and return result.
    source (required)
        yaml as str
    yamlconf (optional)
        yamllint config file to use, if not set will default to a salty version of realaxed.
    """

    if yamlconf is not None:
        conf = YamlLintConfig(file=yamlconf)
    else:
        yamlconf = """
        extends: relaxed
        rules:
          line-length: { max: 256, level: warning }
          empty-lines: disable
          empty-values: {forbid-in-block-mappings: false, forbid-in-flow-mappings: true}
          trailing-spaces: disable
          key-ordering: disable
        """
        conf = YamlLintConfig(yamlconf)

    yaml_out = salt.utils.stringutils.to_str(source)
    problems = []
    for problem in linter.run(yaml_out, conf):
        problems.append(
            {
                "line": problem.line,
                "column": problem.column,
                "level": problem.level,
                "comment": problem.message,
            }
        )
    output = {"source": yaml_out, "problems": problems}
    return output
